# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from astropy.coordinates import SkyCoord, AltAz
from astropy.table import Table
from astropy.time import Time
from regions import PointSkyRegion
from gammapy.data import (
    GTI,
    DataStore,
    EventList,
    FixedPointingInfo,
    Observation,
    observatory_locations,
)
from gammapy.irf import (
    Background2D,
    Background3D,
    EffectiveAreaTable2D,
    EnergyDispersion2D,
)
from gammapy.makers import WobbleRegionsFinder
from gammapy.makers.utils import (
    _map_spectrum_weight,
    guess_instrument_fov,
    make_counts_off_rad_max,
    make_counts_rad_max,
    make_edisp_kernel_map,
    make_effective_livetime_map,
    make_map_background_irf,
    make_map_exposure_true_energy,
    make_observation_time_map,
    ThetaSquaredTable,
    _get_fov_coord,
    project_irf_on_geom,
    integrate_project_irf_on_geom,
)
from gammapy.maps import HpxGeom, MapAxis, RegionGeom, WcsGeom, WcsNDMap
from gammapy.modeling.models import ConstantSpectralModel
from gammapy.utils.coordinates import FoVAltAzFrame
from gammapy.utils.testing import requires_data
from gammapy.utils.time import time_ref_to_dict
from gammapy.utils.coordinates import FoVICRSFrame


@pytest.fixture(scope="session")
def observations():
    """Example observation list for testing."""
    datastore = DataStore.from_dir("$GAMMAPY_DATA/magic/rad_max/data")
    return datastore.get_observations(required_irf="point-like")[0]


@pytest.fixture(scope="session")
def aeff():
    filename = (
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    return EffectiveAreaTable2D.read(filename, hdu="EFFECTIVE AREA")


def geom(map_type, ebounds):
    axis = MapAxis.from_edges(ebounds, name="energy_true", unit="TeV", interp="log")
    if map_type == "wcs":
        return WcsGeom.create(npix=(4, 3), binsz=2, axes=[axis])
    elif map_type == "hpx":
        return HpxGeom(256, axes=[axis])
    else:
        raise ValueError()


@requires_data()
@pytest.mark.parametrize(
    "pars",
    [
        {
            "geom": geom(map_type="wcs", ebounds=[0.1, 1, 10]),
            "shape": (2, 3, 4),
            "sum": 8.103974e08,
        },
        {
            "geom": geom(map_type="wcs", ebounds=[0.1, 10]),
            "shape": (1, 3, 4),
            "sum": 2.387916e08,
        },
        # TODO: make this work for HPX
        # 'HpxGeom' object has no attribute 'separation'
        # {
        #     'geom': geom(map_type='hpx', ebounds=[0.1, 1, 10]),
        #     'shape': '???',
        #     'sum': '???',
        # },
    ],
)
def test_make_map_exposure_true_energy(aeff, pars):
    m = make_map_exposure_true_energy(
        pointing=SkyCoord(2, 1, unit="deg"),
        livetime="42 s",
        aeff=aeff,
        geom=pars["geom"],
    )

    assert m.data.shape == pars["shape"]
    assert m.unit == "m2 s"
    assert_allclose(m.data.sum(), pars["sum"], rtol=1e-5)


def test_map_spectrum_weight():
    axis = MapAxis.from_edges([0.1, 10, 1000], unit="TeV", name="energy_true")
    expo_map = WcsNDMap.create(npix=10, binsz=1, axes=[axis], unit="m2 s")
    expo_map.data += 1
    spectrum = ConstantSpectralModel(const="42 cm-2 s-1 TeV-1")

    weighted_expo = _map_spectrum_weight(expo_map, spectrum)

    assert weighted_expo.data.shape == (2, 10, 10)
    assert weighted_expo.unit == "m2 s"
    assert_allclose(weighted_expo.data.sum(), 100)


@pytest.fixture(scope="session")
def fixed_pointing_info():
    filename = "$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_110380.fits"
    return FixedPointingInfo.read(filename)


@pytest.fixture(scope="session")
def fixed_pointing_info_aligned():
    # Create Fixed Pointing Info aligned between sky and horizon coordinates
    # (removes rotation in FoV and results in predictable solid angles)
    fixed_icrs = SkyCoord(0 * u.deg, 0 * u.deg, frame="icrs")

    return FixedPointingInfo(fixed_icrs=fixed_icrs)


@pytest.fixture(scope="session")
def bkg_3d():
    filename = (
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    return Background3D.read(filename, hdu="BACKGROUND")


@pytest.fixture(scope="session")
def bkg_2d():
    offset_axis = MapAxis.from_bounds(0, 4, nbin=10, name="offset", unit="deg")
    energy_axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=20)
    bkg_2d = Background2D(axes=[energy_axis, offset_axis], unit="s-1 TeV-1 sr-1")
    coords = bkg_2d.axes.get_coord()
    value = np.exp(-0.5 * (coords["offset"] / (2 * u.deg)) ** 2)
    bkg_2d.data = (value * (coords["energy"] / (1 * u.TeV)) ** -2).to_value("")
    return bkg_2d


def bkg_3d_custom(symmetry="constant", fov_align="RADEC"):
    if symmetry == "constant":
        data = np.ones((2, 3, 3))
        sky_edges = [-3, -1, 1, 3] * u.deg
    elif symmetry == "symmetric":
        data = np.ones((2, 3, 3))
        data[:, 1, 1] *= 2
        sky_edges = [-3, -1, 1, 3] * u.deg
    elif symmetry == "hirez_symmetric":
        data = np.ones((3, 3))
        data[1, 1] *= 2
        data = (  # Upscale 3x3 to 9x9 by simply repeating
            data.repeat(3, axis=0).repeat(3, axis=1)[np.newaxis, ...].repeat(2, axis=0)
        )
        sky_edges = np.linspace(-3, 3, 10) * u.deg
    elif symmetry == "asymmetric":
        data = np.indices((3, 3))[1] + 1
        data = np.stack(2 * [data])
        sky_edges = [-3, -1, 1, 3] * u.deg
    else:
        raise ValueError(f"Unknown value for symmetry: {symmetry}")

    energy_axis = MapAxis.from_energy_edges([0.1, 10, 1000] * u.TeV)
    fov_lon_axis = MapAxis.from_edges(sky_edges, name="fov_lon")
    fov_lat_axis = MapAxis.from_edges(sky_edges, name="fov_lat")
    return Background3D(
        axes=[energy_axis, fov_lon_axis, fov_lat_axis],
        data=data,
        unit=u.Unit("s-1 MeV-1 sr-1"),
        interp_kwargs=dict(bounds_error=False, fill_value=None, values_scale="log"),
        fov_alignment=fov_align,
        # allow extrapolation for symmetry tests
    )


def aeff_custom(energy_axis, upscale=3):
    offset = MapAxis.from_bounds(
        -0.25, 2.75, nbin=6 * upscale, unit="deg", name="offset"
    )
    aeff_vals = np.zeros((energy_axis.nbin, offset.nbin))
    Es = energy_axis.center.value
    # Chosen to roughly match HESS areas
    scales = np.repeat(np.array([62, 62, 59, 51, 42, 34]) * 1e4, upscale)
    cores = np.repeat(np.array([0.48, 0.49, 0.52, 0.65, 1.01, 1.61]), upscale)
    ids = list(range(0, offset.nbin))
    for idx, scale, core in zip(ids, scales, cores):
        aeff_vals[:, idx] = (
            scale * core**2 * (1 / (np.sqrt(Es**2 + core**2)) - 1 / core) ** 2
        )

    return EffectiveAreaTable2D([energy_axis, offset], data=aeff_vals, unit="m2")


@requires_data()
def test_map_background_2d(bkg_2d, fixed_pointing_info):
    axis = MapAxis.from_edges([0.1, 1, 10], name="energy", unit="TeV", interp="log")

    obstime = Time("2020-01-01T20:00:00")
    skydir = fixed_pointing_info.get_icrs(obstime).galactic
    geom = WcsGeom.create(
        npix=(3, 3), binsz=4, axes=[axis], skydir=skydir, frame="galactic"
    )

    bkg = make_map_background_irf(
        pointing=skydir,
        ontime="42 s",
        bkg=bkg_2d,
        geom=geom,
        time_start=obstime,
        fov_rotation_step=1.0 * u.deg,
    )

    assert_allclose(bkg.data[:, 1, 1], [1.869025, 0.186903], rtol=1e-5)

    # Check that function works also passing the FixedPointingInfo
    bkg_fpi = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42 s",
        bkg=bkg_2d,
        geom=geom,
        time_start=obstime,
        fov_rotation_step=1.0 * u.deg,
    )
    assert_allclose(bkg.data, bkg_fpi.data, rtol=1e-5)


def make_map_background_irf_with_symmetry(fpi, symmetry="constant"):
    axis = MapAxis.from_edges([0.1, 1, 10], name="energy", unit="TeV", interp="log")
    obstime = Time("2020-01-01T20:00:00")
    return make_map_background_irf(
        pointing=fpi,
        ontime="42 s",
        bkg=bkg_3d_custom(symmetry),
        geom=WcsGeom.create(npix=(3, 3), binsz=4, axes=[axis], skydir=fpi.fixed_icrs),
        time_start=obstime,
        fov_rotation_step=1.0 * u.deg,
    )


def geom(map_type, ebounds, skydir):
    axis = MapAxis.from_edges(ebounds, name="energy", unit="TeV", interp="log")
    if map_type == "wcs":
        return WcsGeom.create(npix=(4, 3), binsz=2, axes=[axis], skydir=skydir)
    elif map_type == "hpx":
        return HpxGeom(256, axes=[axis])
    else:
        raise ValueError()


@requires_data()
@pytest.mark.parametrize(
    "pars",
    [
        {
            "map_type": "wcs",
            "ebounds": [0.1, 1, 10],
            "shape": (2, 3, 4),
            "sum": 1051.960299,
        },
        {
            "map_type": "wcs",
            "ebounds": [0.1, 10],
            "shape": (1, 3, 4),
            "sum": 1051.960299,
        },
        # TODO: make this work for HPX
        # 'HpxGeom' object has no attribute 'separation'
        # {
        #     'geom': geom(map_type='hpx', ebounds=[0.1, 1, 10]),
        #     'shape': '???',
        #     'sum': '???',
        # },
    ],
)
def test_make_map_background_irf(bkg_3d, pars, fixed_pointing_info):
    m = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42 s",
        bkg=bkg_3d,
        geom=geom(
            map_type=pars["map_type"],
            ebounds=pars["ebounds"],
            skydir=fixed_pointing_info.fixed_icrs,
        ),
        time_start=Time("2020-01-01T20:00"),
        fov_rotation_step=1.0 * u.deg,
        oversampling=10,
    )

    assert m.data.shape == pars["shape"]
    assert m.unit == ""
    assert_allclose(m.data.sum(), pars["sum"], rtol=1e-5)


@requires_data()
def test_make_map_background_irf_constant(fixed_pointing_info_aligned):
    m = make_map_background_irf_with_symmetry(
        fpi=fixed_pointing_info_aligned, symmetry="constant"
    )
    for d in m.data:
        assert_allclose(d[1, :], d[1, 0])  # Constant along lon
        assert_allclose(d[0, 1], d[2, 1])  # Symmetric along lat
        with pytest.raises(AssertionError):
            # Not constant along lat due to changes in
            # solid angle (great circle)
            assert_allclose(d[:, 1], d[0, 1])


@requires_data()
def test_make_map_background_irf_sym(fixed_pointing_info_aligned):
    m = make_map_background_irf_with_symmetry(
        fpi=fixed_pointing_info_aligned, symmetry="symmetric"
    )
    for d in m.data:
        assert_allclose(d[1, 0], d[1, 2], rtol=1e-4)  # Symmetric along lon
        assert_allclose(d[0, 1], d[2, 1], rtol=1e-4)  # Symmetric along lat


@requires_data()
def test_make_map_background_irf_asym(fixed_pointing_info_aligned):
    m = make_map_background_irf_with_symmetry(
        fpi=fixed_pointing_info_aligned, symmetry="asymmetric"
    )
    for d in m.data:
        # TODO:
        #  Dimensions of skymap data are [energy, lat, lon] (and is
        #  represented as [lon, lat, energy] in the api, but the bkg irf
        #  dimensions are currently [energy, lon, lat] - Will be changed in
        #  the future (perhaps when IRFs use the skymaps class)
        assert_allclose(d[1, 0], d[1, 2], rtol=1e-4)  # Symmetric along lon
        with pytest.raises(AssertionError):
            assert_allclose(d[0, 1], d[2, 1], rtol=1e-4)  # Symmetric along lat
        assert_allclose(d[0, 1] * 9, d[2, 1], rtol=1e-4)  # Asymmetric along lat


@requires_data()
def test_make_map_background_irf_skycoord(fixed_pointing_info_aligned):
    axis = MapAxis.from_edges([0.1, 1, 10], name="energy", unit="TeV", interp="log")
    position = fixed_pointing_info_aligned.fixed_icrs
    with pytest.raises(TypeError):
        make_map_background_irf(
            pointing=position,
            ontime="42 s",
            bkg=bkg_3d_custom("asymmetric", "ALTAZ"),
            geom=WcsGeom.create(npix=(3, 3), binsz=4, axes=[axis], skydir=position),
            time_start=Time("2020-01-01T20:00:00"),
            fov_rotation_step=1.0 * u.deg,
        )


@requires_data()
def test_make_map_background_irf_altaz_align(fixed_pointing_info):
    def _get_geom(pnt_info, time):
        axis = MapAxis.from_edges([0.1, 1, 10], name="energy", unit="TeV", interp="log")

        return WcsGeom.create(
            npix=(10, 10),
            binsz=0.1,
            axes=[axis],
            skydir=pnt_info.get_icrs(time),
        )

    obstime = Time("2020-01-01T20:00:00")
    location = observatory_locations["ctao_south"]

    map_long_altaz = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42000 s",
        bkg=bkg_3d_custom("asymmetric", "ALTAZ"),
        geom=_get_geom(fixed_pointing_info, obstime),
        time_start=obstime,
        fov_rotation_step=20.0 * u.deg,
        location=location,
    )

    map_short_altaz = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42 s",
        bkg=bkg_3d_custom("asymmetric", "ALTAZ"),
        geom=_get_geom(fixed_pointing_info, obstime),
        time_start=obstime + "20979 s",
        fov_rotation_step=20.0 * u.deg,
        location=location,
    )
    map_long_radec = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42000 s",
        bkg=bkg_3d_custom("asymmetric", "RADEC"),
        geom=_get_geom(fixed_pointing_info, obstime),
        time_start=obstime,
        fov_rotation_step=20.0 * u.deg,
        location=location,
    )
    map_short_altaz_norotation = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="42 s",
        bkg=bkg_3d_custom("asymmetric", "ALTAZ"),
        geom=_get_geom(fixed_pointing_info, obstime),
        time_start=obstime - 21 * u.s,
        fov_rotation_step=360.0 * u.deg,
        location=location,
    )
    map_altaz_long_norotation = make_map_background_irf(
        pointing=fixed_pointing_info,
        ontime="4200 s",
        bkg=bkg_3d_custom("asymmetric", "ALTAZ"),
        geom=_get_geom(fixed_pointing_info, obstime),
        time_start=obstime - 2100 * u.s,
        fov_rotation_step=360.0 * u.deg,
        location=location,
    )
    # Check that background normalisations are consistent
    assert_allclose(np.sum(map_long_altaz.data), np.sum(map_long_radec.data), rtol=1e-2)
    assert np.isclose(
        np.sum(map_long_altaz.data), 1000 * np.sum(map_short_altaz.data), rtol=1e-2
    )
    # Check that results differ when considering short and long observations with
    # AltAz aligned IRFs
    assert_allclose(
        map_long_altaz.data[0, 0, :4], [252123, 250654, 249086, 247564.0], rtol=1e-2
    )
    assert_allclose(
        map_short_altaz.data[0, 0, :4],
        [260.9476, 258.2620, 255.6040, 252.973375],
        rtol=1e-5,
    )

    # Check that results differ when considering RaDec or AltAz aligned IRFs
    assert_allclose(
        map_long_radec.data[0, 0, :4], [197029, 197029, 197029, 197029.0], rtol=1e-5
    )

    # Check that results are independent of the observation duration with AltAz
    # aligned IRFs when the FoV rotation is ignored
    assert_allclose(
        map_altaz_long_norotation.data,
        map_short_altaz_norotation.data * 100,
        rtol=1e-2,
    )


def test_make_edisp_kernel_map():
    migra = MapAxis.from_edges(np.linspace(0.5, 1.5, 50), unit="", name="migra")
    etrue = MapAxis.from_energy_bounds(0.5, 2, 6, unit="TeV", name="energy_true")
    offset = MapAxis.from_edges(np.linspace(0.0, 2.0, 3), unit="deg", name="offset")
    ereco = MapAxis.from_energy_bounds(0.5, 2, 3, unit="TeV", name="energy")

    edisp = EnergyDispersion2D.from_gauss(
        energy_axis_true=etrue, migra_axis=migra, bias=0, sigma=0.01, offset_axis=offset
    )

    geom = WcsGeom.create(10, binsz=0.5, axes=[ereco, etrue])
    pointing = SkyCoord(0, 0, frame="icrs", unit="deg")
    edispmap = make_edisp_kernel_map(edisp, pointing, geom)

    kernel = edispmap.get_edisp_kernel(position=pointing)
    assert_allclose(kernel.pdf_matrix[:, 0], (1.0, 1.0, 0.0, 0.0, 0.0, 0.0), atol=1e-14)
    assert_allclose(kernel.pdf_matrix[:, 1], (0.0, 0.0, 1.0, 1.0, 0.0, 0.0), atol=1e-14)
    assert_allclose(kernel.pdf_matrix[:, 2], (0.0, 0.0, 0.0, 0.0, 1.0, 1.0), atol=1e-14)


@requires_data()
def test_make_counts_rad_max(observations):
    pos = SkyCoord(083.6331144560900, +22.0144871383400, unit="deg", frame="icrs")
    on_region = PointSkyRegion(pos)
    energy_axis = MapAxis.from_energy_bounds(
        0.05, 100, nbin=6, unit="TeV", name="energy"
    )
    geom = RegionGeom.create(region=on_region, axes=[energy_axis])
    counts = make_counts_rad_max(geom, observations.rad_max, observations.events)

    assert_allclose(np.squeeze(counts.data), np.array([547, 188, 52, 8, 0, 0]))


@requires_data()
def test_make_counts_off_rad_max(observations):
    pos = SkyCoord(83.6331, +22.0145, unit="deg", frame="icrs")
    on_region = PointSkyRegion(pos)
    energy_axis = MapAxis.from_energy_bounds(
        0.05, 100, nbin=6, unit="TeV", name="energy"
    )

    region_finder = WobbleRegionsFinder(n_off_regions=3)
    region_off, wcs = region_finder.run(on_region, pos)
    geom_off = RegionGeom.from_regions(regions=region_off, axes=[energy_axis], wcs=wcs)

    counts_off = make_counts_off_rad_max(
        geom_off=geom_off, rad_max=observations.rad_max, events=observations.events
    )

    assert_allclose(np.squeeze(counts_off.data), np.array([1641, 564, 156, 24, 0, 0]))


class TestTheta2Table:
    def setup_class(self):
        self.observations = []
        for sign in [-1, 1]:
            events = Table()
            events["RA"] = [0.0, 0.0, 0.0, 0.0, 10.0] * u.deg
            events["DEC"] = sign * ([0.0, 0.05, 0.9, 10.0, 10.0] * u.deg)
            events["ENERGY"] = [1.0, 1.0, 1.5, 1.5, 10.0] * u.TeV
            events["OFFSET"] = [0.1, 0.1, 0.5, 1.0, 1.5] * u.deg
            events["TIME"] = Time("2025-01-01") + [0.1, 0.2, 0.3, 0.4, 0.5] * u.s

            obs_info = dict(
                OBS_ID=0,
                DEADC=1,
                GEOLON=16.500222222222224,
                GEOLAT=-23.271777777777775,
                ALTITUDE=1834.9999999997833,
            )

            meta = time_ref_to_dict("2010-01-01")
            obs_info.update(meta)
            events.meta.update(obs_info)
            gti = GTI.create(
                start=[1] * u.s,
                stop=[3] * u.s,
                reference_time=Time("2010-01-01", scale="tt"),
            )
            pointing = FixedPointingInfo(
                fixed_icrs=SkyCoord(0 * u.deg, sign * 0.5 * u.deg),
            )

            self.observations.append(
                Observation(
                    events=EventList(events),
                    gti=gti,
                    pointing=pointing,
                )
            )

    @pytest.mark.parametrize(
        "case_config",
        [
            # CASE 1: Standard (single observation)
            {
                "obs_idx": [0],
                "axis_bounds": (0, 0.2),
                "kwargs": {},
                "expected": {
                    "counts": [2, 0, 0, 0],
                    "counts_off": [1, 0, 0, 0],
                    "acceptance": [1, 1, 1, 1],
                    "alpha": [1, 1, 1, 1],
                },
            },
            # CASE 2: Two observations (identical in this setup)
            {
                "obs_idx": [0, 1],  # Assuming self.observations has at least 2 items
                "axis_bounds": (0, 0.2),
                "kwargs": {},
                "expected": {
                    "counts": [4, 0, 0, 0],
                    "counts_off": [2, 0, 0, 0],
                    "acceptance": [2, 2, 2, 2],
                    "alpha": [1, 1, 1, 1],
                },
            },
            # CASE 3: Energy selection
            {
                "obs_idx": [0],
                "axis_bounds": (0, 0.2),
                "kwargs": {"energy_edges": [1.0, 11] * u.TeV},
                "expected": {
                    "counts": [2, 0, 0, 0],
                    "counts_off": [1, 0, 0, 0],
                    "acceptance": [1, 1, 1, 1],
                    "alpha": [1, 1, 1, 1],
                    "energy_meta": [1.0, 11] * u.TeV,
                },
            },
            # CASE 4: Multiple OFF regions
            {
                "obs_idx": [0],
                "axis_bounds": (0, 0.19),
                "kwargs": {"energy_edges": [1.0, 11] * u.TeV, "off_regions_number": 2},
                "expected": {
                    "counts": [2, 0, 0, 0],
                    "counts_off": [0, 0, 0, 0],
                    "alpha": [0.5, 0.5, 0.5, 0.5],
                    "excess": [2, 0, 0, 0],
                    "sqrt_ts": [2.0962941, 0.0, 0.0, 0.0],
                    "energy_meta": [1.0, 11] * u.TeV,
                },
            },
            # CASE 5: Overlapping multiple OFF regions
            {
                "obs_idx": [0],
                "axis_bounds": (0, 0.22),
                "kwargs": {"energy_edges": [1.0, 11] * u.TeV, "off_regions_number": 2},
                "expected": {
                    "counts": [2, 0, 0, 0],
                    "counts_off": [1, 0, 0, 0],
                    "alpha": [1, 1, 1, 1],
                    "excess": [1, 0, 0, 0],
                    "sqrt_ts": [0.582922, 0.0, 0.0, 0.0],
                    "energy_meta": [1.0, 11] * u.TeV,
                },
            },
            # CASE 6: User-defined OFF region
            {
                "obs_idx": [0],
                "axis_bounds": (0, 0.1),
                "kwargs": {
                    "position_off": SkyCoord(0, -1.0, frame="icrs", unit="deg"),
                    "energy_edges": [1.2, 11] * u.TeV,
                    "off_regions_number": 1,
                },
                "expected": {
                    "counts": [0, 0, 0, 0],
                    "counts_off": [1, 0, 0, 0],
                    "alpha": [1, 1, 1, 1],
                    "excess": [-1, 0, 0, 0],
                    "sqrt_ts": [-1.17741, 0.0, 0.0, 0.0],
                    "energy_meta": [1.2, 11] * u.TeV,
                },
            },
        ],
    )
    def test_theta_squared_table_values(self, case_config):
        position = SkyCoord(ra=0, dec=0, unit="deg", frame="icrs")
        b_min, b_max = case_config["axis_bounds"]
        axis = MapAxis.from_bounds(b_min, b_max, nbin=4, interp="lin", unit="deg2")

        observations = [self.observations[i] for i in case_config["obs_idx"]]

        theta2_maker = ThetaSquaredTable(
            observations=observations,
            position=position,
            theta_squared_axis=axis,
            **case_config["kwargs"],
        )

        table = theta2_maker.run()

        assert len(table) == 4
        assert table["theta2_min"].unit == "deg2"
        assert_allclose(table.meta["ON_RA"], 0 * u.deg)
        assert_allclose(table.meta["ON_DEC"], 0 * u.deg)

        exp = case_config["expected"]
        if "counts" in exp:
            assert_allclose(table["counts"], exp["counts"])
        if "counts_off" in exp:
            assert_allclose(table["counts_off"], exp["counts_off"])
        if "alpha" in exp:
            assert_allclose(table["alpha"], exp["alpha"])
        if "acceptance" in exp:
            assert_allclose(table["acceptance"], exp["acceptance"])

        if "excess" in exp:
            assert_allclose(table["excess"], exp["excess"])
        if "sqrt_ts" in exp:
            assert_allclose(table["sqrt_ts"], exp["sqrt_ts"])

        if "energy_meta" in exp:
            assert_allclose(table.meta["Energy_filter"], exp["energy_meta"])

    @pytest.mark.parametrize(
        "case_params",
        [
            # ERROR 1: Wrong axis unit (deg instead of deg2)
            {
                "axis_unit": "deg",
                "axis_bounds": (0, 0.2),
                "kwargs": {},
                "error_msg": "The theta2 axis should be equivalent to deg2",
            },
            # ERROR 2: OFF position is same as ON position
            {
                "axis_unit": "deg2",
                "axis_bounds": (0, 0.2),
                "kwargs": {"position_off_is_on": True},
                "error_msg": "The specified OFF region overlaps with the ON region. This is currently forbidden. To fix this, either choose another OFF position or reduce the region radius.",
            },
            # ERROR 3: Incorrect energy edges (3 values instead of 2)
            {
                "axis_unit": "deg2",
                "axis_bounds": (0, 0.2),
                "kwargs": {"energy_edges": [1.2, 11, 20] * u.TeV},
                "error_msg": "Only supports one energy interval but 2 passed.",
            },
            # ERROR 4: Region radius larger than offset (axis bounds too large)
            {
                "axis_unit": "deg2",
                "axis_bounds": (0, 0.3),
                "kwargs": {"energy_edges": [1.0, 11] * u.TeV, "off_regions_number": 1},
                "error_msg": "The ON region radius is larger than its separation from observation pointing position. This will cause the ON and OFF regions to overlap, which is not permitted. Reduce the ON region radius to at least 0.5 deg.",
            },
            # ERROR 5: User-defined OFF position AND multiple OFF regions requested (inconsistent)
            {
                "axis_unit": "deg2",
                "axis_bounds": (0, 0.05),
                "kwargs": {
                    "position_off": SkyCoord(0, -1.0, frame="icrs", unit="deg"),
                    "energy_edges": [1.2, 11] * u.TeV,
                    "off_regions_number": 2,
                },
                "error_msg": "If ``off_regions_number`` is larger than 1, you cannot provide a fixed OFF position. Instead set ``position_off`` to be None.",
            },
        ],
    )
    def test_theta_squared_table_errors(self, case_params):
        position = SkyCoord(ra=0, dec=0, unit="deg", frame="icrs")

        b_min, b_max = case_params["axis_bounds"]
        unit = case_params["axis_unit"]
        axis = MapAxis.from_bounds(b_min, b_max, nbin=4, interp="lin", unit=unit)

        kwargs = case_params["kwargs"].copy()

        if kwargs.pop("position_off_is_on", False):
            kwargs["position_off"] = position

        error_message = case_params["error_msg"]
        with pytest.raises(ValueError, match=error_message):
            table_maker = ThetaSquaredTable(
                observations=[self.observations[0]],
                position=position,
                theta_squared_axis=axis,
                **kwargs,
            )
            table_maker.run()


@requires_data()
def test_guess_instrument_fov(observations):
    with pytest.raises(ValueError):
        guess_instrument_fov(observations)

    ds = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    obs_hess = ds.obs(23523)

    assert_allclose(guess_instrument_fov(obs_hess), 2.5 * u.deg)

    obs_no_aeff = obs_hess.copy(in_memory=True, aeff=None)
    with pytest.raises(ValueError):
        guess_instrument_fov(obs_no_aeff)


@requires_data()
def test_make_observation_time_map():
    ds = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    obs_id = ds.obs_table["OBS_ID"][ds.obs_table["OBJECT"] == "MSH 15-5-02"][:3]
    observations = ds.get_observations(obs_id)
    source_pos = SkyCoord(228.32, -59.08, unit="deg")
    geom = WcsGeom.create(
        skydir=source_pos,
        binsz=0.02,
        width=(6, 6),
        frame="icrs",
        proj="CAR",
    )
    obs_time = make_observation_time_map(observations, geom, offset_max=2.5 * u.deg)
    obs_time_center = obs_time.get_by_coord(source_pos)
    assert_allclose(obs_time_center, 1.2847, rtol=1e-3)
    assert obs_time.unit == u.hr


@requires_data()
def test_make_effective_livetime_map():
    ds = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    obs_id = ds.obs_table["OBS_ID"][ds.obs_table["OBJECT"] == "MSH 15-5-02"][:3]
    observations = ds.get_observations(obs_id)
    source_pos = SkyCoord(228.32, -59.08, unit="deg")
    offset_pos = SkyCoord(322.00, 0.1, unit="deg", frame="galactic")

    energy_axis_true = MapAxis.from_energy_bounds(
        10 * u.GeV, 1 * u.TeV, nbin=2, name="energy_true"
    )
    geom = WcsGeom.create(
        skydir=source_pos,
        binsz=0.02,
        width=(6, 6),
        frame="galactic",
        proj="CAR",
        axes=[energy_axis_true],
    )
    obs_time = make_effective_livetime_map(observations, geom, offset_max=2.5 * u.deg)
    obs_time_center = obs_time.get_by_coord((source_pos, energy_axis_true.center))
    assert_allclose(obs_time_center, [0, 1.2847], rtol=1e-3)

    obs_time_offset = obs_time.get_by_coord((offset_pos, energy_axis_true.center))
    assert_allclose(obs_time_offset, [0, 0.242814], rtol=1e-3)

    assert obs_time.unit == u.hr


@requires_data()
def test_project_irf(aeff):
    ebounds = [0.1, 1, 10]
    axis = MapAxis.from_edges(ebounds, name="energy_true", unit="TeV", interp="log")
    geom = WcsGeom.create(npix=(4, 3), binsz=2, axes=[axis])
    pointing = SkyCoord(2, 1, unit="deg")
    fov_frame = FoVICRSFrame(origin=pointing)
    proj_irf_geom = project_irf_on_geom(geom, aeff, fov_frame)

    assert geom.data_shape == proj_irf_geom.data.shape
    assert_allclose(
        proj_irf_geom.data[:, 1, 1], [373700.92608812, 2477509.59635073], rtol=1e-5
    )
    assert proj_irf_geom.geom.center_skydir == geom.center_skydir


@requires_data()
def test_integrate_project_irf(bkg_2d, fixed_pointing_info):
    axis = MapAxis.from_edges([0.1, 1, 10], name="energy", unit="TeV", interp="log")
    obstime = Time("2020-01-01T20:00:00")
    skydir = fixed_pointing_info.get_icrs(obstime).galactic
    geom = WcsGeom.create(
        npix=(3, 3), binsz=4, axes=[axis], skydir=skydir, frame="galactic"
    )
    fov_frame = FoVICRSFrame(origin=skydir)
    int_proj_irf_geom = integrate_project_irf_on_geom(geom, bkg_2d, fov_frame)

    assert geom.data_shape == int_proj_irf_geom.data.shape
    assert_allclose(int_proj_irf_geom.data[:, 1, 1], [0.0445006, 0.00445006], rtol=1e-5)
    assert int_proj_irf_geom.geom.center_skydir == geom.center_skydir


@pytest.mark.filterwarnings("ignore:.*Angular separation .*direction")
def test_get_fov_coords():
    crab = SkyCoord(83.63333333, 22.01444444, unit="deg", frame="icrs")
    location = observatory_locations.get("ctao_north")
    time_start = Time("2025-01-01T00:00:00")

    fov_origin = crab.transform_to(AltAz(location=location, obstime=time_start))
    fov_frame = FoVAltAzFrame(origin=fov_origin, location=location, obstime=time_start)

    # Check that pixel centers end up at correct offsets
    # (Uses FOV frame centered on the Crab)
    center_sep = 0.5  # separation between square pixel centers
    # is the pixel size
    sky_geom = WcsGeom.create(npix=(3, 3), binsz=center_sep, skydir=crab, proj="TAN")
    sky_coord = sky_geom.to_image().get_coord().skycoord
    coords = _get_fov_coord(sky_coord, fov_frame, use_offset=True)["offset"]
    assert_allclose(0, coords[1, 1].value, atol=1e-11)
    assert_allclose(center_sep, coords[0, 1].value, rtol=3e-5)
    assert_allclose(center_sep, coords[1, 0].value, rtol=3e-5)

    center_sep = 0.01
    sky_geom = WcsGeom.create(npix=(3, 3), binsz=center_sep, skydir=crab, proj="TAN")
    sky_coord = sky_geom.to_image().get_coord().skycoord
    coords = _get_fov_coord(sky_coord, fov_frame, use_offset=True)["offset"]
    assert_allclose(center_sep, coords[0, 1].value, rtol=3e-5)
    assert_allclose(center_sep, coords[1, 0].value, rtol=3e-5)

    # Check that pixel centers end up at correct offsets
    # after letting the sky drift away from the starting position
    fov_frame = FoVAltAzFrame(
        origin=fov_origin, location=location, obstime=time_start + 4.303333 * u.minute
    )
    center_sep = 0.5
    sky_geom = WcsGeom.create(npix=(3, 3), binsz=center_sep, skydir=crab, proj="TAN")
    sky_coord = sky_geom.to_image().get_coord().skycoord
    coords = _get_fov_coord(sky_coord, fov_frame, use_offset=True)["offset"]
    assert_allclose(center_sep, coords[1, 0].value, rtol=3e-5)
    assert_allclose(2 * center_sep, coords[1, 1].value, rtol=3e-5)
    assert_allclose(3 * center_sep, coords[1, 2].value, rtol=5e-5)

    # Check that pixel centers end up at correct offsets
    # after letting the sky drift away from the starting position
    obs_times = time_start + np.linspace(0, 4.303333, 2) * u.minute
    fov_frame = FoVAltAzFrame(origin=fov_origin, location=location, obstime=obs_times)
    center_sep = 0.5
    sky_geom = WcsGeom.create(npix=(3, 3), binsz=center_sep, skydir=crab, proj="TAN")
    sky_coord = sky_geom.to_image().get_coord().skycoord
    coords = _get_fov_coord(sky_coord[..., np.newaxis], fov_frame, use_offset=True)[
        "offset"
    ]
    assert coords.unit == u.deg
    assert list(coords.shape) == [2, 3, 3]
    assert_allclose(0, coords[0, 1, 1].value, atol=1e-11)
    assert_allclose(center_sep, coords[0, 1, 0].value, rtol=3e-5)
    assert_allclose(center_sep, coords[0, 0, 1].value, rtol=3e-5)
    assert_allclose(2 * center_sep, coords[1, 1, 1].value, rtol=3e-5)
    assert_allclose(3 * center_sep, coords[1, 1, 2].value, rtol=5e-5)


def test_project_irf_on_geom():
    location = observatory_locations.get("ctao_north")
    crab = SkyCoord(83.63333333, 22.01444444, unit="deg", frame="icrs")

    # Test projection to geom aligned with the FOV binning and orientation
    obstime = Time("2025-01-01T00:00:00")
    origin = crab.transform_to(AltAz(location=location, obstime=obstime))
    fov_frame = FoVAltAzFrame(origin=origin, location=location, obstime=obstime)

    axis = MapAxis.from_energy_bounds(
        energy_min=0.1, energy_max=10, nbin=2, name="energy_true", unit="TeV"
    )
    sky_geom = WcsGeom.create(
        npix=(3, 3), binsz=0.5, axes=[axis], skydir=crab, proj="TAN"
    )
    aeff = aeff_custom(axis)
    sky_irf = project_irf_on_geom(sky_geom, aeff, fov_frame)

    ref0 = aeff.evaluate(offset=0 * u.deg).flatten().value
    ref1 = aeff.evaluate(offset=sky_geom.separation(crab)[0, 1]).flatten().value

    assert_allclose(ref0, sky_irf.data[:, 1, 1])
    assert_allclose(ref1, sky_irf.data[:, 0, 1])
    assert_allclose(ref1, sky_irf.data[:, 1, 0])

    obs_times = obstime - np.linspace(0, 2.15, 2) * u.minute

    fov_frame = FoVAltAzFrame(origin=origin, location=location, obstime=obs_times)
    aeff = aeff_custom(axis, upscale=5)
    sky_irf = project_irf_on_geom(sky_geom, aeff, fov_frame)

    ref3 = (ref0 + ref1) / 2
    assert_allclose(ref3, sky_irf.data[:, 1, 1])


def test_integrate_project_irf_on_geom():
    location = observatory_locations.get("ctao_north")
    crab = SkyCoord(83.63333333, 22.01444444, unit="deg", frame="icrs")

    # Test projection to geom aligned with the FOV binning and orientation
    obstime = Time("2025-01-01T00:04:00")
    origin = crab.transform_to(AltAz(location=location, obstime=obstime))
    axis = MapAxis.from_edges([0.1, 1.1, 11.1], name="energy", unit="TeV", interp="log")
    sky_geom = WcsGeom.create(
        npix=(3, 3), binsz=2, axes=[axis], skydir=crab, proj="TAN"
    )

    fov_frame = FoVAltAzFrame(origin=origin, location=location, obstime=obstime)
    bkg_irf = bkg_3d_custom(symmetry="asymmetric")
    bkg_sky = integrate_project_irf_on_geom(sky_geom, bkg_irf, fov_frame)

    ref1 = (
        1
        * u.TeV
        * bkg_irf.evaluate(fov_lon=0 * u.deg, fov_lat=0 * u.deg, energy=1 * u.TeV)
        * sky_geom.solid_angle()[0, 1, 1]
    )
    ref2 = (
        10
        * u.TeV
        * bkg_irf.evaluate(fov_lon=0 * u.deg, fov_lat=2 * u.deg, energy=1 * u.TeV)
        * sky_geom.solid_angle()[0, 1, 2]
    )

    assert_allclose(bkg_sky.data[0, 1, 1], ref1.value, rtol=1e-9)
    assert_allclose(bkg_sky.data[1, 2, 1], ref2.value, rtol=1e-3)

    assert bkg_sky.unit.is_equivalent(1 / u.s)

    bkg_irf = bkg_3d_custom(symmetry="hirez_symmetric")
    bkg_sky = integrate_project_irf_on_geom(sky_geom, bkg_irf, fov_frame)
    ref3 = (  # center center pixel, value = 2
        1
        * u.TeV
        * bkg_irf.evaluate(fov_lon=0 * u.deg, fov_lat=0 * u.deg, energy=1 * u.TeV)
    ).value
    ref4 = (  # right edge center pixel, value = 1
        1
        * u.TeV
        * bkg_irf.evaluate(fov_lon=0 * u.deg, fov_lat=2 * u.deg, energy=1 * u.TeV)
    ).value
    omega11 = sky_geom.solid_angle()[0, 1, 1].value  # center center pixel
    omega12 = sky_geom.solid_angle()[0, 1, 2].value  # right edge center pixel
    omega21 = sky_geom.solid_angle()[0, 2, 1].value  # bottom edge center pixel
    assert_allclose(bkg_sky.data[0, 1, 1], ref3 * omega11, rtol=1e-9)
    assert_allclose(bkg_sky.data[0, 1, 2], ref4 * omega12, rtol=1e-9)
    # Times chosen such that at second step the IRF FoV has moved
    # over one full geom pixel
    obs_times = obstime - np.linspace(0, 8.6, 2) * u.minute
    fov_timedep = FoVAltAzFrame(origin=origin, location=location, obstime=obs_times)
    bkg_sky = integrate_project_irf_on_geom(sky_geom, bkg_irf, fov_timedep)

    assert_allclose(
        (ref3 + ref4) / 2 * omega11,
        bkg_sky.data[0, 1, 1],
    )
    assert_allclose(
        (ref3 + ref4) / 2 * omega12,
        bkg_sky.data[0, 1, 2],
    )
    with pytest.raises(AssertionError):
        assert_allclose(
            (ref3 + ref4) / 2 * omega21,
            bkg_sky.data[0, 2, 1],
        )
    assert_allclose(
        (ref4 + ref4) / 2 * omega21,
        bkg_sky.data[0, 2, 1],
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from numpy.testing import assert_allclose
from gammapy.datasets import Datasets
from gammapy.datasets import SpectrumDatasetOnOff

from gammapy.modeling.models import (
    SkyModel,
    LogUniformPrior,
)
import pytest
from gammapy.modeling.sampler import Sampler
from gammapy.utils.testing import requires_data, requires_dependency
from gammapy.modeling.models import Models
from gammapy.modeling.selection import BayesianModelSelection


@pytest.fixture()
def alternative_models():
    alternative_models = {}
    model_type = "lp"
    prior_types = ["uninformative", "strong"]

    for prior_type in prior_types:
        model_name = f"{model_type}({prior_type})"
        model = SkyModel.create(spectral_model=model_type, name=f"crab-{model_name}")
        model.spectral_model.reference.value = 1
        model.spectral_model.alpha.frozen = True
        model.spectral_model.beta.frozen = True

        if "uninformative" in prior_type:
            model.spectral_model.amplitude.prior = LogUniformPrior(min=1e-14, max=1e-8)

        elif "strong" in prior_type:
            model.spectral_model.amplitude.prior = LogUniformPrior(min=1e-11, max=1e-10)

        alternative_models[model_name] = Models(model)

    return alternative_models


@pytest.fixture()
def bms_results(alternative_models):
    path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
    datasets = Datasets()
    for id in ["23523", "23526", "23559", "23592"]:
        dataset = SpectrumDatasetOnOff.read(f"{path}pha_obs{id}.fits")
        datasets.append(dataset)

    sampler_opts = {
        "live_points": 50,
        "frac_remain": 0.5,
        "log_dir": None,
    }

    sampler = Sampler(backend="ultranest", sampler_opts=sampler_opts)

    bms = BayesianModelSelection(
        sampler, posterior_downsample_factor=5, n_prior_samples=10
    )

    bms_results = bms.run(datasets, alternative_models)

    return bms_results


@requires_data()
@requires_dependency("arviz")
def test_bayesian_model_selection(bms_results):
    inference_result = bms_results["lp(uninformative)"]
    assert "Statistics summary in deviance scale" in str(inference_result)

    assert_allclose(inference_result.elpd_loo.elpd_loo, 218.91, rtol=1e-1)

    assert_allclose(inference_result.priors["amplitude"].entropy(), -22.7, rtol=1e-1)
    bms_results["lp(strong)"].priors["amplitude"].entropy()

    aic = -2 * inference_result.logl + 2 * inference_result.dof
    assert_allclose(inference_result.aic, aic)

    parameter_table = inference_result.parameters_table()
    assert parameter_table["parameter"] == "amplitude"
    assert_allclose(parameter_table["median"], 5.49e-11, rtol=1e-1)
    assert_allclose(parameter_table["mean"], 5.50e-11, rtol=1e-1)
    assert_allclose(parameter_table["mode"], 5.52e-11, rtol=1e-1)
    assert_allclose(parameter_table["value at max ln(L)"], 5.44e-11, rtol=1e-1)

    stats_table = bms_results.stats_table()
    assert_allclose(stats_table["-2logl"], 216.356, rtol=1e-1)
    assert bms_results["lp(uninformative)"].logz < bms_results["lp(strong)"].logz

    table = bms_results.stats_difference_table("lp(uninformative)")
    test_str = "H0: lp(uninformative) - H1: lp(strong)"
    assert table["Model (prior)"][0] == test_str
    assert_allclose(table["-2logl"], 0, atol=1e-5)


@pytest.mark.xfail
@requires_data()
@requires_dependency("arviz")
def test_prior_sensitivity_table(bms_results):
    bms_results.prior_sensitivity_table()
    psense = bms_results["lp(uninformative)"].prior_sensitivity_table()
    assert_allclose(psense.prior.amplitude, 0.01, rtol=5e-1)
    assert_allclose(psense.likelihood.amplitude, 0.097, rtol=5e-1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.astro.darkmatter import (
    DarkMatterAnnihilationSpectralModel,
    DarkMatterDecaySpectralModel,
    PrimaryFlux,
)
from gammapy.modeling.models import Models, SkyModel
from gammapy.utils.testing import assert_quantity_allclose, requires_data


@requires_data()
def test_primary_flux():
    with pytest.raises(ValueError):
        PrimaryFlux(channel="Spam", mDM=1 * u.TeV)

    primflux = PrimaryFlux(channel="W", mDM=1 * u.TeV)
    actual = primflux(500 * u.GeV)
    desired = 9.3319318e-05 / u.GeV
    assert_quantity_allclose(actual, desired)


@pytest.mark.parametrize(
    "mass, expected_flux, source, expected_exception",
    [
        (1.6, 0.00025037, "pppc4", None),
        (11, 0.00549445, "cosmixs", None),
        (75, None, "nonexistend", ValueError),
        (75, None, "pppc4", ValueError),
    ],
)
@requires_data()
def test_primary_flux_interpolation(mass, expected_flux, source, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            PrimaryFlux(channel="aZ", mDM=mass * u.TeV, source=source)
        return
    primflux = PrimaryFlux(channel="W", mDM=mass * u.TeV, source=source)
    actual = primflux(500 * u.GeV)
    assert_quantity_allclose(actual, expected_flux / u.GeV, rtol=1e-5)


@pytest.mark.parametrize(
    "model_class, jfactor_unit, expected_flux, expected_dnde, source",
    [
        (
            DarkMatterAnnihilationSpectralModel,
            "GeV2 cm-5",
            6.19575457e-14,
            2.97831615e-16,
            None,
        ),
        (
            DarkMatterDecaySpectralModel,
            "GeV cm-2",
            3.209234e-2,
            2.33485775e-05,
            "pppc4",
        ),
        (
            DarkMatterAnnihilationSpectralModel,
            "GeV2 cm-5",
            6.03197683e-14,
            3.52065879e-16,
            "cosmixs",
        ),
        (DarkMatterDecaySpectralModel, "GeV cm-2", 0.031677, 2.771877e-05, "cosmixs"),
    ],
)
@requires_data()
def test_dm_spectral_model(
    tmpdir, jfactor_unit, model_class, expected_flux, expected_dnde, source
):
    channel = "b"
    mass = 5 * u.TeV
    jfactor = 3.41e19 * u.Unit(jfactor_unit)
    energy_min = 0.01 * u.TeV
    energy_max = 10 * u.TeV

    model = model_class(mass=mass, channel=channel, jfactor=jfactor, source=source)
    flux = model.integral(energy_min=energy_min, energy_max=energy_max).to("cm-2 s-1")
    dnde = model.evaluate(energy=1 * u.TeV, scale=1).to("cm-2 s-1 TeV-1")

    sky_model = SkyModel(
        spectral_model=model,
        name="skymodel",
    )
    models = Models([sky_model])
    filename = tmpdir / "model.yaml"
    models.write(filename, overwrite=True)
    new_models = Models.read(filename)

    assert_quantity_allclose(flux.value, expected_flux, rtol=1e-3)
    assert_quantity_allclose(dnde.value, expected_dnde, rtol=1e-3)

    assert new_models[0].spectral_model.channel == model.channel
    assert new_models[0].spectral_model.z == model.z
    assert_allclose(new_models[0].spectral_model.jfactor.value, model.jfactor.value)
    assert new_models[0].spectral_model.mass.value == 5
    assert new_models[0].spectral_model.mass.unit == u.TeV


@requires_data()
def test_primary_flux_cosmixs():
    with pytest.raises(ValueError):
        PrimaryFlux(channel="Spam", mDM=1 * u.TeV)

    primflux = PrimaryFlux(channel="W", mDM=1 * u.TeV, source="cosmixs")
    actual = primflux(500 * u.GeV)
    desired = 0.00013085 / u.GeV
    assert_quantity_allclose(actual, desired, rtol=1e-4)

    with pytest.raises(ValueError):
        PrimaryFlux(channel="q", mDM=1 * u.TeV, source="cosmixs")
    with pytest.raises(ValueError):
        PrimaryFlux(channel="V->e", mDM=1 * u.TeV, source="cosmixs")
    with pytest.raises(ValueError):
        PrimaryFlux(channel="V->mu", mDM=1 * u.TeV, source="cosmixs")
    with pytest.raises(ValueError):
        PrimaryFlux(channel="V->tau", mDM=1 * u.TeV, source="cosmixs")

    with pytest.raises(ValueError):
        PrimaryFlux(channel="d", mDM=1 * u.TeV, source="pppc4")
    with pytest.raises(ValueError):
        PrimaryFlux(channel="u", mDM=1 * u.TeV, source="pppc4")
    with pytest.raises(ValueError):
        PrimaryFlux(channel="s", mDM=1 * u.TeV, source="pppc4")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for FluxCollectionEstimator
"""

import numpy as np
import astropy.units as u
import pytest


from gammapy.maps import Map, WcsGeom, MapAxis
from gammapy.modeling import Fit, Sampler
from gammapy.modeling.models import (
    Models,
    SkyModel,
    PowerLawSpectralModel,
    FoVBackgroundModel,
)
from gammapy.datasets import MapDataset, Datasets

from gammapy.estimators.points.sed import FluxCollectionEstimator


@pytest.fixture
def simple_dataset():
    axis = MapAxis.from_energy_bounds(0.1, 10, 1, unit="TeV")
    geom = WcsGeom.create(npix=20, binsz=0.02, axes=[axis])
    simple_dataset = MapDataset.create(geom)
    simple_dataset.mask_safe += np.ones(simple_dataset.data_shape, dtype=bool)
    simple_dataset.counts += 2
    simple_dataset.background += 1

    spectral = PowerLawSpectralModel(index=2.0, amplitude="1e-12 cm-2 s-1 TeV-1")
    simple_dataset.models = [
        SkyModel(spectral_model=spectral, name="test-source"),
        FoVBackgroundModel(dataset_name=simple_dataset.name),
    ]

    return simple_dataset


@pytest.fixture
def energy_edges():
    return np.array([1.0, 3.0, 10.0]) * u.TeV


class MockFit(Fit):
    """Mock Fit for multiple models with distinct values."""

    def run(self, datasets):
        # assign distinct best-fit norms for each model i
        for ds in datasets:
            for i, ev in enumerate(ds.evaluators.values()):
                m = ev.model
                if hasattr(m.spectral_model, "norm"):
                    m.spectral_model.norm.value = 1.0 + i
                    m.spectral_model.norm.error = 0.1 + 0.05 * i
        return {"success": True}

    def confidence(self, datasets, parameter, sigma):
        return {"errn": 0.1 * sigma, "errp": 0.2 * sigma}


@pytest.fixture
def mock_fit():
    return MockFit()


class MockSamplerResult:
    def __init__(self, ns):
        n_samples = 200
        # points: shape (n_samples, ns)
        rng = np.random.default_rng(42)
        points = 1.0 + np.arange(ns) * 0.5 + 0.1 * rng.standard_normal((n_samples, ns))
        weights = np.ones(n_samples) / n_samples

        self.sampler_results = {
            "weighted_samples": {
                "points": points,
                "weights": weights,
            }
        }


class MockSamplerMulti(Sampler):
    """Mock sampler for multiple-source."""

    def __init__(self, ns=1):
        super().__init__(backend="mock", sampler_opts={})
        self.ns = ns

    def run(self, datasets):
        return MockSamplerResult(self.ns)


@pytest.fixture
def mock_sampler_multi():
    return MockSamplerMulti(ns=2)


@pytest.fixture
def mock_sampler():
    return MockSamplerMulti(ns=1)


def test_init(simple_dataset, energy_edges):
    model = simple_dataset.models["test-source"]
    est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[model],
    )
    assert isinstance(est.solver, Sampler)


def test_run(simple_dataset, energy_edges, mock_fit):
    model = simple_dataset.models["test-source"]
    est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[model],
        solver=mock_fit,
        reoptimize=True,
    )

    result = est.run(Datasets([simple_dataset]))
    fp = result["flux_points"]["test-source"]

    assert len(fp["dnde"].data) == 2
    assert fp["dnde"].unit == u.Unit("cm-2 s-1 TeV-1")
    assert np.all(np.isfinite(fp["dnde"]))
    assert np.all(fp["ts"] >= 0)


def test_asymmetric_errors_present(simple_dataset, energy_edges, mock_fit):
    model = simple_dataset.models["test-source"]
    est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[model],
        solver=mock_fit,
        selection_optional=["errn-errp"],
    )

    result = est.run(Datasets([simple_dataset]))
    fp = result["flux_points"]["test-source"]
    assert np.all(np.isfinite(fp["dnde_errn"]))
    assert np.all(np.isfinite(fp["dnde_errp"]))


def test_inconsistent_geometry_raises(simple_dataset, mock_fit):
    # dataset with different geom
    geom2 = WcsGeom.create(npix=(3, 3), binsz=0.1)
    ds2 = simple_dataset.copy(name="bad-ds")
    ds2.counts = Map.from_geom(geom2, data=np.ones(geom2.data_shape))

    model = simple_dataset.models["test-source"]
    est = FluxCollectionEstimator(
        energy_edges=[1, 3, 10] * u.TeV,
        models=[model],
        solver=mock_fit,
    )

    with pytest.raises(ValueError):
        est.run(Datasets([simple_dataset, ds2]))


def test_run_sampler_case(simple_dataset, energy_edges, mock_sampler):
    model = simple_dataset.models["test-source"]

    est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[model],
        solver=mock_sampler,
    )

    result = est.run(Datasets([simple_dataset]))
    fp = result["flux_points"]["test-source"]

    assert len(fp["dnde"].data) == 2
    assert np.all(np.isfinite(fp["dnde"]))
    assert np.all(np.isfinite(fp["dnde_errn"]))
    assert np.all(np.isfinite(fp["dnde_errp"]))
    assert np.all(np.isfinite(fp["dnde_ul"]))
    assert np.all(np.isfinite(fp["ts"]))

    # sampler-only data
    samples = result["samples"]["dnde"]
    assert "test-source" in samples
    assert len(samples["test-source"]) == 2


def test_run_multi_source(simple_dataset, energy_edges, mock_fit, mock_sampler_multi):
    # Add two sources
    m1 = simple_dataset.models["test-source"]
    m2 = m1.copy(name="test-source-2")
    simple_dataset.models = Models([m1, m2])

    fit_est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[m1, m2],
        solver=mock_fit,
        selection_optional=["errn-errp"],
    )

    sampler_est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[m1, m2],
        solver=mock_sampler_multi,
    )

    for est in [fit_est, sampler_est]:
        result = est.run(Datasets([simple_dataset]))
        flux_points = result["flux_points"]
        nbin = len(energy_edges) - 1

        assert set(flux_points.keys()) == {"test-source", "test-source-2"}

        for name in ["test-source", "test-source-2"]:
            fp = flux_points[name]
            assert len(fp["dnde"].data) == nbin
            assert np.all(np.isfinite(fp["dnde"]))
            assert np.all(np.isfinite(fp["dnde_errn"]))
            assert np.all(np.isfinite(fp["dnde_errp"]))
            assert np.all(np.isfinite(fp["dnde_ul"]))
            assert np.all(np.isfinite(fp["ts"]))
            assert np.all(fp["ts"] >= 0)

    assert result["solver_results"].shape == (nbin,)

    samples = result["samples"]["dnde"]
    assert "test-source" in samples
    assert "test-source-2" in samples
    assert len(samples["test-source"]) == nbin
    assert samples["test-source"][0].shape[0] == 200


def test_run_multi_dataset(simple_dataset, energy_edges, mock_fit):
    ds2 = simple_dataset.copy(name="dataset-2")
    ds2.counts = ds2.counts.copy(data=12 * np.ones(ds2.counts.data.shape))
    ds2.background = ds2.background.copy(data=9 * np.ones(ds2.background.data.shape))

    datasets = Datasets([simple_dataset, ds2])

    m1 = simple_dataset.models["test-source"]
    m2 = m1.copy(name="test-source-2")

    models = Models([m1, m2])
    simple_dataset.models = models
    ds2.models = models

    est = FluxCollectionEstimator(
        energy_edges=energy_edges,
        models=[m1, m2],
        solver=mock_fit,
        selection_optional=["errn-errp"],
        reoptimize=True,
    )

    result = est.run(datasets)
    flux_points = result["flux_points"]
    nbin = len(energy_edges) - 1

    for name in ["test-source", "test-source-2"]:
        fp = flux_points[name]
        assert len(fp["dnde"].data) == nbin
        assert np.all(np.isfinite(fp["dnde"]))
        assert np.all(np.isfinite(fp["ts"]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from numpy.testing import assert_allclose
from astropy.time import Time, TimeDelta
import astropy.units as u
from gammapy.utils.time import (
    absolute_time,
    extract_time_info,
    time_to_fits,
    time_to_fits_header,
    time_ref_from_dict,
    time_ref_to_dict,
    time_relative_to_ref,
    unique_time_info,
)


def test_time_to_fits():
    time = Time("2001-01-01T00:00:00")
    fits_time_simple = time_to_fits(time)

    epoch = Time(10000, format="mjd", scale="tt")

    fits_time_epoch = time_to_fits(time, epoch)
    fits_time_unit = time_to_fits(time, unit=u.d)

    assert_allclose(fits_time_simple, 4.485024e09 * u.s)
    assert_allclose(fits_time_epoch, 3.621024e09 * u.s)
    assert_allclose(fits_time_unit, 51910.000743 * u.d)


def test_time_to_fits_header():
    time = Time("2001-01-01T00:00:00")
    fits_header_value, fits_header_unit = time_to_fits_header(time)
    assert_allclose(fits_header_value, 4.485024e09)
    assert fits_header_unit == "s"


def test_time_ref_from_dict():
    d = dict(MJDREFI=51910, MJDREFF=0.00074287036841269583)
    mjd = d["MJDREFF"] + d["MJDREFI"]

    time = time_ref_from_dict(d)
    assert time.format == "mjd"
    assert time.scale == "tt"
    assert_allclose(time.mjd, mjd)


def test_time_ref_to_dict():
    time = Time("2001-01-01T00:00:00")

    d = time_ref_to_dict(time)

    assert set(d) == {"MJDREFI", "MJDREFF", "TIMESYS"}
    assert d["MJDREFI"] == 51910
    assert_allclose(d["MJDREFF"], 0.00074287036841269583)
    assert d["TIMESYS"] == "tt"


def test_time_relative_to_ref():
    time_ref_dict = dict(MJDREFI=500, MJDREFF=0.5)
    time_ref = time_ref_from_dict(time_ref_dict)
    delta_time_1sec = TimeDelta(1.0, format="sec")
    time = time_ref + delta_time_1sec

    delta_time = time_relative_to_ref(time, time_ref_dict)

    assert_allclose(delta_time.sec, delta_time_1sec.sec)


def test_absolute_time():
    time_ref_dict = dict(MJDREFI=51000, MJDREFF=0.5)
    time_ref = time_ref_from_dict(time_ref_dict)
    delta_time_1sec = TimeDelta(1.0, format="sec")
    time = time_ref + delta_time_1sec

    abs_time = absolute_time(delta_time_1sec, time_ref_dict)

    assert abs_time.value == time.utc.isot


def test_extract_time_info():
    dd = dict(MJDREFI=1, MJDREFF=2, TIMEUNIT=3, TIMESYS=4, TIMEREF=5, TELESCOPE="IACT")
    time_row = extract_time_info(dd)

    assert time_row["TIMESYS"] == 4


def test_check_time_info():
    rows = [
        dict(MJDREFI=1, MJDREFF=2, TIMEUNIT=3, TIMESYS=4, TIMEREF=5),
        dict(MJDREFI=1, MJDREFF=2, TIMEUNIT=5, TIMESYS=4, TIMEREF=5),
    ]

    assert unique_time_info(rows) is False



"""
Make a theta-square plot
========================

This tutorial explains how to make such a plot, that is the distribution
of event counts as a function of the squared angular distance, to a test
position.

"""


######################################################################
# Setup
# -----
#

# %matplotlib inline
import matplotlib.pyplot as plt
from astropy.coordinates import SkyCoord
from astropy import units as u
from gammapy.data import DataStore
from gammapy.maps import MapAxis
from gammapy.makers.utils import ThetaSquaredTable
from gammapy.visualization import plot_theta_squared_table


######################################################################
# Get some data
# -------------
#
# Some data taken on the Crab by H.E.S.S. are used.
#

data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
observations = data_store.get_observations([23523, 23526])


######################################################################
# Define a test position
# -----------------------
#
# Here we define the position of Crab
#

position = SkyCoord(83.6324, 22.0174, unit="deg")
print(position)


######################################################################
# Creation of the theta2 plot
# ---------------------------
#
# By default, the distribution of the OFF counts in squared angular distance is
# calculated from the mirror reflected coordinates of the test position, assuming
# therefore a single OFF position. However, one can set manually both the
# coordinates of the ``off_position``. It is worth to note that overlapping
# regions are always forbidden to avoid correlated OFF counts, therefore the
# user should take care in the choice of the ``off_position``.
#

separation = position.separation(observations[0].pointing.fixed_icrs)
position_angle = 180 * u.deg
off_position = position.directional_offset_by(
    position_angle=position_angle, separation=separation * 2
)
print(off_position)

# %%
theta2_axis = MapAxis.from_bounds(0, 0.2, nbin=20, interp="lin", unit="deg2")
theta2_table_maker = ThetaSquaredTable(
    observations=observations,
    position=position,
    theta_squared_axis=theta2_axis,
    position_off=off_position,
)
theta2_table = theta2_table_maker.run()

plt.figure(figsize=(10, 5))
plot_theta_squared_table(theta2_table)
plt.show()

######################################################################
# Alternatively, it can be requested a number of reflected OFF positions that will
# be calculated through the `~gammapy.makers.WobbleRegionsFinder`. As mentioned before,
# the user should be cautious that the regions (ON and OFF) do not overlap, otherwise
# only the mirror reflected region will be adopted as OFF.
#

off_regions_number = 3
theta2_axis = MapAxis.from_bounds(0, 0.1, nbin=20, interp="lin", unit="deg2")
theta2_table_maker_offreg = ThetaSquaredTable(
    observations=observations,
    position=position,
    theta_squared_axis=theta2_axis,
    off_regions_number=off_regions_number,
)
theta2_table_offreg = theta2_table_maker_offreg.run()

plt.figure(figsize=(10, 5))
plot_theta_squared_table(theta2_table_offreg)
plt.show()


######################################################################
# Making a theta2 plot for a given energy range
# ---------------------------------------------
#
# With the function `~gammapy.makers.utils.ThetaSquaredTable`, one can
# also select a fixed energy range.
#

theta2_table_maker_en = ThetaSquaredTable(
    observations=observations,
    position=position,
    theta_squared_axis=theta2_axis,
    energy_edges=[1.2, 11] * u.TeV,
)
theta2_table_en = theta2_table_maker_en.run()

plt.figure(figsize=(10, 5))
plot_theta_squared_table(theta2_table_en)
plt.show()


######################################################################
# Statistical significance of a detection
# ---------------------------------------
#
# To get the significance of a signal, the usual method consists of using the reflected background method
# (see the maker tutorial: :doc:`/user-guide/makers/reflected`) to compute the WStat statistics
# (see :ref:`wstat`, :doc:`/user-guide/stats/fit_statistics`). This is the well-known method of [LiMa1983]_
# using ON and OFF regions.
#
# The following tutorials show how to get an excess significance:
#
# -  :doc:`/tutorials/analysis-1d/spectral_analysis`
# -  :doc:`/tutorials/analysis-1d/extended_source_spectral_analysis`
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""classes containing the workflow steps supported by the high level interface"""

import collections.abc
import numpy as np
from gammapy.utils.scripts import make_name, requires_module

try:
    import ray
except ImportError:
    ray = None  # Fallback if ray is not available


class Product:
    """Represents a product generated during a processing step in a workflow."""

    def __init__(self, name=None, step_name=None, data=None):
        """
        Initialize a Product instance.

        Parameters
        ----------
        name : str, optional
            A label or identifier for the product.
            Default is None, and the name is randomly generated.
        step_name : str, optional
            The name of the processing step that produced this product.
            Default is None.
        data : object, optional
            The data associated with the product, it can be any object.
        """

        if step_name is None:
            step_name = ""
        self.step_name = step_name
        self.data = data
        if name is None:
            name = make_name()
        self.name = name
        self.pid = make_name()  # unique id generated for each product

    @property
    @requires_module("ray")
    def needs_update(self):
        """
        Check whether the product's data needs to be updated.
        Only available if `ray` is available.

        Returns
        -------
        bool
            True if `data` is a `ray.ObjectRef`, indicating that the data is a reference
            to a remote object and may need to be resolved or updated.
        """
        return isinstance(self.data, ray.ObjectRef)

    def copy(self):
        """
        Create a shallow copy of the product with a new unique identifier.

        Returns
        -------
        Product
            A new `Product` instance with the same `tag`, `step_name`, and `data`,
            but a newly generated `name`.
        """
        return self.__class__(name=self.name, step_name=self.step_name, data=self.data)


class Products(collections.abc.MutableSequence):
    """
    Container for managing a list of `Product` instances generated by the workflow.

    This class provides list-like behavior and additional utilities for
    selecting, copying, and updating products, including support for
    Ray-based asynchronous data resolution.

    Parameters
    ----------
    products : list of `Product`, optional
        Initial list of products. Defaults to an empty list.
    """

    def __init__(self, products=None):
        if products is None:
            products = []
        self._products = products

    @property
    def names(self):
        """List of product names."""
        return [p.name for p in self._products]

    @property
    def pids(self):
        """List of product names."""
        return [p.pid for p in self._products]

    @property
    def unique_names(self):
        """List of unique product names. Return formatted as step_name.product_name.product_id."""
        names = []
        for p in self._products:
            components = [p.step_name, p.name, p.pid]
            name = ".".join(components)
            names.append(name)
        return names

    @property
    def data(self):
        """List of `Product` data."""
        return [p.data for p in self._products]

    def index(self, key):
        """
        Get the index of a product by pid or position.

        Parameters
        ----------
        key : int, slice, or str
            Index or name of the product.

        Returns
        -------
        int or slice
            Index corresponding to the key.
        """
        if isinstance(key, (int, slice)):
            return key
        elif isinstance(key, str):
            return self.names.index(key)

    def copy(self):
        """Create a deep copy of the products container."""
        return self.__class__([p.copy() for p in self])

    def __len__(self):
        return len(self._products)

    def __getitem__(self, key):
        if isinstance(key, np.ndarray) and key.dtype == bool:
            return self.__class__(list(np.array(self._products)[key]))
        else:
            return self._products[self.index(key)]

    def __delitem__(self, key):
        del self._products[self.index(key)]

    def insert(self, idx, model):
        self._products.insert(idx, model)

    def __setitem__(self, key, product):
        if isinstance(product, Product):
            if isinstance(key, np.ndarray) and key.dtype == bool:
                ind = np.nonzero(key)[0]
                for idx in ind:
                    self._products[int(idx)] = product
            else:
                self._products[self.index(key)] = product
        else:
            raise TypeError(f"Invalid type: {product!r}")

    def select_from_list(self, requirements):
        """
        Select products based on a list of criteria.

        Parameters
        ----------
        requirements : list of dict
            List of selection criteria dictionaries passed to select method.

        Returns
        -------
        Products
            A new `Products` instance with selected products.
        """
        selection = Products([])
        for req in requirements:
            selection.extend(self.select(**req))
        return selection

    def select(self, name=None, step_name=None, pid=None):
        """
        Select products matching the given criteria.

        Parameters
        ----------
        name : str, optional
            Name to match. Default is None.
            Default is None, and name filter is not applied.
        step_name : str, optional
            Step name to match.
            Default is None, and `step_name` filter is not applied.
        pid : str, optional
            Product pid to match.
            Default is None, and `pid` filter is not applied.

        Returns
        -------
        Products
            A new `Products` instance with matching products.
        """

        mask = self.selection_mask(name=name, step_name=step_name, pid=pid)
        return self[mask]

    def selection_mask(self, name=None, step_name=None, pid=None):
        """
        Create a boolean mask for selected products.

        Parameters
        ----------
        name : str, optional
            Names to match.
            Default is None, and name filter is not applied.
        step_name : str, optional
            Step names to match.
            Default is None, and `step_name` filter is not applied.
        pid : str, optional
            Product pid to match.
            Default is None, and `pid` filter is not applied.

        Returns
        -------
        mask : np.ndarray
            Boolean array indicating selected products.
        """

        mask = np.ones(len(self), dtype=bool)
        for idx, p in enumerate(self._products):
            if name is not None:
                mask[idx] &= p.name == name
            if step_name is not None:
                mask[idx] &= p.step_name == step_name
            if pid is not None:
                mask[idx] &= p.pid == pid
        return mask

    @requires_module("ray")
    def get(self):
        """
        Resolve and update products whose data is only a reference (`ray.ObjectRef`).
        Only available if `ray` is available.

        This method fetches remote data using `ray.get()` and updates
        the corresponding products in-place.

        """
        ind = np.nonzero([p.needs_update for p in self._products])[0]
        refs_to_run = {data for data in np.array(self.data)[ind]}  # unique objects
        if refs_to_run:
            results = ray.get(list(refs_to_run))
            products = [item for sublist in results for item in sublist]  # flatten
            for product in products:
                mask = self.selection_mask(pid=product.pid)
                self[mask] = product



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from gammapy.utils.interpolation import LogScale
from gammapy.utils.testing import assert_allclose


def test_logscale_adaptive_behavior():
    """
    Test that LogScale adapts its clipping threshold based on input dtype.
    """
    log_scale = LogScale()

    # Case 1: float64 input (High Precision)
    # -----------------------------------------------------------
    # 1e-40 is well within float64 limits (~1e-308), so it should NOT be clipped.
    values_64 = np.array([1, 1e-5, 1e-40], dtype=np.float64)
    log_values_64 = log_scale(values_64)

    # Expectation: actual log(1e-40), not log(1e-38)
    # This verifies the fix for the "Silent Bias" issue.
    assert_allclose(log_values_64, np.log(values_64))

    # Inverse check: Should recover 1e-40 exactly, not 0.
    inv_values_64 = log_scale.inverse(log_values_64)
    assert_allclose(inv_values_64, values_64)

    # Case 2: float32 input (Standard Precision)
    # -----------------------------------------------------------
    # 0 is smaller than float32 tiny (~1.17e-38), so it SHOULD be clipped
    # to maintain numerical stability for 32-bit workflows.
    tiny_32 = np.finfo(np.float32).tiny
    values_32 = np.array([1, 1e-5, 0], dtype=np.float32)

    log_values_32 = log_scale(values_32)

    # Expectation: 0 is clipped to tiny_32
    expected_clipped_32 = np.array([1, 1e-5, tiny_32], dtype=np.float32)
    assert_allclose(log_values_32, np.log(expected_clipped_32))

    # Inverse check: Under the original LogScale semantics, values close to tiny
    # are mapped back to 0.
    inv_values_32 = log_scale.inverse(log_values_32)
    expected_inv_32 = np.array([1, 1e-5, 0], dtype=np.float32)
    assert_allclose(inv_values_32, expected_inv_32, rtol=1e-6)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
from scipy.stats import gaussian_kde, norm


def plot_samples_violin_vs_energy(
    energy_edges,
    samples_per_band,
    weights_per_band=None,
    energy_power=0,
    bw_method="scott",
    grid_size=200,
    ax=None,
    errorbar_kwargs=None,
    errorbar_ul_kwargs=None,
    violin_kwargs=None,
    violin_clip=None,
    y_label="dN/dE",
):
    """Plot sample violin distributions per energy bin on log–log axes.

    Unlike standard error bars which only shows summary statistics,
    a violin plot shows the full probability density for each energy bin.

    This function draws one violin per energy interval, using weighted
    kernel-density estimation (KDE) in log-space. The median and
    1σ-equivalent quantiles are overlaid using Gammapy-style flux-point
    error bars. Users can optionally apply an ``E_center**p`` scaling to the
    flux values and clip violin tails to a specified containment interval.
    Samples of other energy-dependent parameters can also be plotted,
    in that case the ``y_label`` should be changed accordingly.

    Parameters
    ----------
    energy_edges : `~numpy.ndarray` or `~astropy.units.Quantity`
        Monotonically increasing energy bin edges of length ``nbins + 1``.
        Must be positive and finite for log scaling.
    samples_per_band : list of `~astropy.units.Quantity`
        List of sample arrays, one per energy bin. Each array contains
        draws from the flux posterior (or other distribution) in that bin.
    weights_per_band : list of `~numpy.ndarray`, optional
        Per-sample weights for each bin, same length as ``samples_per_band``.
        If omitted, uniform weights are assumed.
    energy_power : float, optional
        Power of energy to multiply y-axis with. Default is 0.
    bw_method : str or float, optional
        Bandwidth selection passed to `~scipy.stats.gaussian_kde`.
        Default is "scott".
    grid_size : int, optional
        Number of evaluation points in log-flux space for the KDE grid.
        Default is 200.
    ax : `~matplotlib.axes.Axes`, optional
        Matplotlib axes to draw the violins and error bars. Default is None.
    errorbar_kwargs : dict, optional
        Keyword arguments forwarded to `~matplotlib.pyplot.errorbar` for the quantile bars.
    errorbar_ul_kwargs : dict, optional
        Keyword arguments forwarded to `~matplotlib.pyplot.errorbar` for the upper limit bars.
    violin_kwargs : dict, optional
        Keyword arguments forwarded to `~matplotlib.axes.Axes.fill` for violin plot.
    violin_clip : (float, float), optional
        Lower and upper containment fractions (in [0, 1]) used to clip
        the violin tails in log-space. If omitted, defaults to
        ``(norm.cdf(-4), norm.cdf(4))``.
    y_label : str, optional
        Base label for the y-axis, before unit and scaling prefixes.
        Default is "dN/dE".

    Returns
    -------
    ax : `~matplotlib.axes.Axes`
        Matplotlib axes.

    Notes
    -----
    - KDE is performed on the transformed variable ``log10(E^p * F)``,
      where ``p = energy_power``. The violin polygon is constructed by
      mirroring the normalized log-density around the bin center in
      log-energy.
    - Violin clipping is applied only to the KDE grid, not to the
      quantile-based error bars.
    - Negative flux samples are replaced by a small positive surrogate
      (10% of the smallest positive sample) to allow log-space evaluation.
    - The plotted error bars correspond to the 16%, 50%, and 84%
      weighted percentiles of the *unclipped* samples, unless the median
      is non-positive, in which case an upper limit marker is drawn
      at the 98% percentile. For a gaussian distribution these values corresponds to
      1σ errors and 2σ upper limit plotted by default by the `~gammapy.estimators.FluxPoints.plot` method.
    """
    if ax is None:
        ax = plt.gca()

    ax.set_xscale("log")
    ax.set_yscale("log")

    # Energy edges
    energy_edges = u.Quantity(energy_edges)
    unit_x = f"{energy_edges.unit}"
    edges = np.asarray(energy_edges.to_value())
    _validate_energy_edges(edges)

    unit_y = f"{samples_per_band[0].unit}"
    nbins = len(edges) - 1

    samples_per_band, weights_per_band = _validate_inputs(
        samples_per_band, weights_per_band, nbins
    )

    # clip defaults
    if violin_clip is None:
        violin_clip = (norm.cdf(-4), norm.cdf(4))
    lc, uc = violin_clip
    if not (0 <= lc < uc <= 1):
        raise ValueError("violin_clip must satisfy 0 <= low < high <= 1.")

    # Errorbar styles
    errorbar_kwargs = errorbar_kwargs or {}
    errorbar_kwargs_defaults = dict(
        marker="o",
        ms=4.5,
        mec="black",
        mfc="white",
        color="black",
        capsize=2.5,
        elinewidth=1.2,
        lw=1.2,
    )
    for key in errorbar_kwargs_defaults.keys():
        errorbar_kwargs.setdefault(key, errorbar_kwargs_defaults[key])

    errorbar_ul_kwargs = errorbar_ul_kwargs or {}
    errorbar_ul_kwargs_defaults = dict(
        marker="v",
        ms=7,
        mec="black",
        mfc="white",
        color="black",
        capsize=2.5,
        elinewidth=1.2,
        lw=1.2,
    )
    for key in errorbar_ul_kwargs_defaults.keys():
        errorbar_ul_kwargs.setdefault(key, errorbar_ul_kwargs_defaults[key])

    if violin_kwargs is None:
        violin_kwargs = dict()
    violin_kwargs_defaults = dict(
        alpha=0.45,
        color="C0",
        edgecolor="black",
        lw=0.8,
    )
    for key in violin_kwargs_defaults.keys():
        violin_kwargs.setdefault(key, violin_kwargs_defaults[key])

    quantiles = [100 * norm.cdf(-1), 50, 100 * norm.cdf(1), 100 * norm.cdf(2)]

    # Precompute bin geometry
    emins = edges[:-1]
    emaxs = edges[1:]
    xlog_min = np.log10(emins)
    xlog_max = np.log10(emaxs)

    centers = 0.5 * (xlog_min + xlog_max)
    halfwidths = 0.5 * (xlog_max - xlog_min) * 0.99

    for samples, weights, xlog_c, hwlog, emin, emax in zip(
        samples_per_band, weights_per_band, centers, halfwidths, emins, emaxs
    ):
        samples = samples.to_value(unit_y)
        s, smin_pos = _sanitize_samples(samples)
        w = _sanitize_weights(weights, len(s))
        if w is None or s.size == 0:
            continue

        # transform
        ylog, scale = _compute_log_transformed(s, xlog_c, energy_power)

        # KDE
        grid_full, dens_full = _kde_evaluate(ylog, w, grid_size, bw_method)

        # clipping interval
        y_high = np.percentile(s, uc * 100, weights=w, method="inverted_cdf") * scale
        y_low = np.percentile(s, lc * 100, weights=w, method="inverted_cdf") * scale
        y_low = max(smin_pos * scale, y_low)

        if y_high > y_low > 0:
            ymin = max(grid_full.min(), np.log10(y_low))
            ymax = min(grid_full.max(), np.log10(y_high))
            mask = (grid_full >= ymin) & (grid_full <= ymax)
            ygrid_log, dens = grid_full[mask], dens_full[mask]
        else:
            ygrid_log, dens = grid_full, dens_full

        # violin polygon
        _draw_violin(ax, xlog_c, hwlog, ygrid_log, dens, violin_kwargs)

        # quantile bars
        _draw_errorbar(
            ax,
            xlog_c,
            emin,
            emax,
            samples,
            weights,
            quantiles,
            scale,
            errorbar_kwargs,
            errorbar_ul_kwargs,
        )

    # Labels
    ax.set_xlabel(f"Energy [{unit_x}]")

    if energy_power:
        p = f"{energy_power:g}"
        unit_str = rf"[{unit_x}$^{p}$ × {unit_y}]"
        ax.set_ylabel(rf"$E^{p}\,\times$ {y_label} {unit_str}")
    else:
        ax.set_ylabel(f"{y_label} [{unit_y}]")

    return ax


def _validate_inputs(samples_per_band, weights_per_band, nbins):
    """Validate that samples and weights."""
    if len(samples_per_band) != nbins:
        raise ValueError("samples_per_band must match number of bins.")

    if weights_per_band is None:
        weights_per_band = [None] * nbins
    elif len(weights_per_band) != nbins:
        raise ValueError("weights_per_band must match number of bins.")
    return samples_per_band, weights_per_band


def _validate_energy_edges(edges):
    """Validate that energy_edges is 1D, positive, finite, and strictly increasing."""
    if edges.ndim != 1 or edges.size < 2:
        raise ValueError("energy_edges must be 1D and contain at least two values.")
    if (
        not np.all(np.isfinite(edges))
        or np.any(edges <= 0)
        or not np.all(np.diff(edges) > 0)
    ):
        raise ValueError("energy_edges must be strictly increasing, finite, and > 0.")


def _sanitize_samples(samples):
    """Return finite samples with negative values adjusted for log-scale use."""
    s = samples.copy()
    mask = np.isfinite(s)
    s = s[mask]

    s_pos = s[s > 0]
    if np.any(s < 0):
        if s_pos.size > 0:
            s[s < 0] = 0.1 * s_pos.min()

    return s, s_pos.min()


def _sanitize_weights(weights, size):
    """Return normalised finite non-negative weights or None if empty."""
    if weights is None:
        w = np.ones(size, float)
    else:
        w = np.asarray(weights.copy(), float)
        w = w[np.isfinite(w) & (w >= 0)]

    total = w.sum()
    return w / total if total > 0 else None


def _compute_log_transformed(s, xlog_center, energy_power):
    """Transform samples to log10(E^p × F) and return log-values and linear scale."""
    if energy_power:
        shift = energy_power * xlog_center
        scale = (10**xlog_center) ** energy_power
    else:
        shift = 0.0
        scale = 1.0

    return np.log10(s) + shift, scale


def _kde_evaluate(ylog, w, grid_size, bw_method):
    """Evaluate weighted KDE on a fixed log-grid and normalise density."""
    grid = np.linspace(ylog.min(), ylog.max(), int(grid_size))
    kde = gaussian_kde(ylog, weights=w, bw_method=bw_method)
    dens = kde(grid)
    peak = dens.max()
    return grid, dens / peak if peak > 0 else dens


def _draw_violin(ax, x_center, hwlog, ygrid_log, dens, violin_kwargs):
    """Draw a symmetric violin polygon in log-space."""
    xlog_left = x_center - hwlog * dens
    xlog_right = x_center + hwlog * dens

    x_left = 10**xlog_left
    x_right = 10**xlog_right
    y_lin = 10**ygrid_log

    x_poly = np.concatenate([x_left, x_right[::-1]])
    y_poly = np.concatenate([y_lin, y_lin[::-1]])

    return ax.fill(x_poly, y_poly, zorder=2, **violin_kwargs)


def _draw_errorbar(ax, xlog_c, emin, emax, s, w, quantiles, scale, err_kw, err_ul_kw):
    """Draw Gammapy-style flux-point error bars or upper limits."""
    y_qs = [
        np.percentile(s, q, weights=w, method="inverted_cdf") * scale for q in quantiles
    ]

    y_lo, y_med, y_hi, y_ul = y_qs[0], y_qs[1], y_qs[2], y_qs[3]
    if y_med > 0:
        yerr = np.array([[max(0.0, y_med - y_lo)], [max(0.0, y_hi - y_med)]])
        kwargs = err_kw
    else:
        yerr = None
        y_med = y_ul
        kwargs = err_ul_kw
    x_c = 10 ** (xlog_c)
    xerr = np.array([[x_c - emin], [emax - x_c]])
    return ax.errorbar(x_c, y_med, xerr=xerr, yerr=yerr, **(kwargs or {}), zorder=5)



r"""
.. _logparabola2-spectral-model:

Log parabola 2 spectral model
=============================

This model parametrises a log parabola spectrum,
where the energy scale of the exponent, :math:`E_{\mathrm{s}}`, and the reference energy, :math:`E_0`, can be different.
It is defined by the following equation:

.. math::
        \phi(E) = \phi_0 \left( \frac{E}{E_0} \right) ^ {
          - \alpha - \beta \log{ \left( \frac{E}{E_{\mathrm{s}}} \right) }
        }

Note that :math:`\log` refers to the natural logarithm.
If you have parametrization based on :math:`\log_{10}` you can use the
:func:`~gammapy.modeling.models.LogParabola2SpectralModel.from_log10` method.

"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.modeling.models import LogParabola2SpectralModel, Models, SkyModel

energy_bounds = [0.1, 100] * u.TeV
model = LogParabola2SpectralModel(
    alpha=2.3,
    amplitude="1e-12 cm-2 s-1 TeV-1",
    reference=10 * u.TeV,
    beta=0.5,
    escale=1 * u.TeV,
)
model.plot(energy_bounds)
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(spectral_model=model, name="log-parabola2-model")
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .core import FluxPoints
from .lightcurve import LightCurveEstimator
from .profile import FluxProfileEstimator
from .sed import FluxPointsEstimator, FluxCollectionEstimator
from .sensitivity import SensitivityEstimator

__all__ = [
    "FluxCollectionEstimator",
    "FluxPoints",
    "FluxPointsEstimator",
    "FluxProfileEstimator",
    "LightCurveEstimator",
    "SensitivityEstimator",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from astropy import units as u
from astropy.visualization import quantity_support
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from gammapy.maps.axes import UNIT_STRING_FORMAT
from gammapy.utils.array import array_stats_str
from gammapy.visualization.utils import add_colorbar
from ..core import IRF


class PSF(IRF):
    """PSF base class."""

    def normalize(self):
        """Normalise PSF."""
        super().normalize(axis_name="rad")

    def containment(self, rad, **kwargs):
        """Containment tof the PSF at given axes coordinates.

        Parameters
        ----------
        rad : `~astropy.units.Quantity`
            Rad value.
        **kwargs : dict
            Other coordinates.
            You can view the available options through "psf.axes.names".

        Returns
        -------
        containment : `~numpy.ndarray`
            Containment.
        """
        containment = self.integral(axis_name="rad", rad=rad, **kwargs)
        return containment.to("")

    def containment_radius(self, fraction, factor=20, **kwargs):
        """Containment radius at given axes coordinates.

        Parameters
        ----------
        fraction : float or `~numpy.ndarray`
            Containment fraction.
        factor : int, optional
            Up-sampling factor of the rad axis, determines the precision of the
            computed containment radius.
            Default is 20.
        **kwargs : dict
            Other coordinates.
            You can view the available options through "psf.axes.names".

        Returns
        -------
        radius : `~astropy.coordinates.Angle`
            Containment radius.

        Examples
        --------
        >>> from gammapy.data import DataStore
        >>> import astropy.units as u
        >>> data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
        >>> obs = data_store.obs(23523)
        >>> containment_radius = obs.psf.containment_radius(energy_true=0.1*u.TeV ,offset=0.3*u.deg, fraction=0.68)
        """
        # TODO: this uses a lot of numpy broadcasting tricks, maybe simplify...
        from gammapy.datasets.map import RAD_AXIS_DEFAULT

        output = np.broadcast(*kwargs.values(), fraction)

        try:
            rad_axis = self.axes["rad"]
        except KeyError:
            rad_axis = RAD_AXIS_DEFAULT

        # upsample for better precision
        rad = rad_axis.upsample(factor=factor).center

        axis = tuple(range(output.ndim))
        rad = np.expand_dims(rad, axis=axis).T
        containment = self.containment(rad=rad, **kwargs)

        fraction_idx = np.argmin(np.abs(containment - fraction), axis=0)
        return rad[fraction_idx].reshape(output.shape)

    def info(
        self,
        fraction=(0.68, 0.95),
        energy_true=([1.0], [10.0]) * u.TeV,
        offset=0 * u.deg,
    ):
        """
        Print PSF summary information.

        The containment radius for given fraction, energies and thetas is
        computed and printed on the command line.

        Parameters
        ----------
        fraction : list, optional
            Containment fraction to compute containment radius for, between 0 and 1.
            Default is (0.68, 0.95).
        energy_true : `~astropy.units.Quantity`, optional
            Energies to compute containment radius for.
            Default is ([1.0], [10.0]) TeV.
        offset : `~astropy.units.Quantity`, optional
            Offset to compute containment radius for.
            Default is 0 deg.

        Returns
        -------
        info : string
            Formatted string containing the summary information.
        """
        info = "\nSummary PSF info\n"
        info += "----------------\n"
        info += array_stats_str(self.axes["offset"].center.to("deg"), "Theta")
        info += array_stats_str(self.axes["energy_true"].edges[1:], "Energy hi")
        info += array_stats_str(self.axes["energy_true"].edges[:-1], "Energy lo")

        containment_radius = self.containment_radius(
            energy_true=energy_true, offset=offset, fraction=fraction
        )

        energy_true, offset, fraction = np.broadcast_arrays(
            energy_true, offset, fraction, subok=True
        )

        for idx in np.ndindex(containment_radius.shape):
            info += f"{100 * fraction[idx]:.2f} containment radius "
            info += f"at offset = {offset[idx]} "
            info += f"and energy_true = {energy_true[idx]:4.1f}: "
            info += f"{containment_radius[idx]:.3f}\n"

        return info

    def plot_containment_radius_vs_energy(
        self, ax=None, fraction=(0.68, 0.95), offset=(0, 1) * u.deg, **kwargs
    ):
        """Plot containment fraction as a function of energy.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes. Default is None.
        fraction : list of float or `~numpy.ndarray`, optional
            Containment fraction between 0 and 1.
            Default is (0.68, 0.95).
        offset : `~astropy.units.Quantity`, optional
            Offset array.
            Default is (0, 1) deg.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.plot`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
             Matplotlib axes.

        """
        ax = plt.gca() if ax is None else ax

        energy_true = self.axes["energy_true"]

        with quantity_support():
            for theta in offset:
                for frac in fraction:
                    plot_kwargs = kwargs.copy()
                    radius = self.containment_radius(
                        energy_true=energy_true.center, offset=theta, fraction=frac
                    )
                    plot_kwargs.setdefault("label", f"{theta}, {100 * frac:.1f}%")
                    ax.plot(energy_true.center, radius, **plot_kwargs)

        energy_true.format_plot_xaxis(ax=ax)
        ax.legend(loc="best")
        ax.set_ylabel(
            f"Containment radius [{ax.yaxis.units.to_string(UNIT_STRING_FORMAT)}]"
        )
        ax.yaxis.set_major_formatter(mtick.FormatStrFormatter("%.2f"))
        return ax

    def plot_containment_radius(
        self,
        ax=None,
        fraction=0.68,
        add_cbar=True,
        axes_loc=None,
        kwargs_colorbar=None,
        **kwargs,
    ):
        """Plot containment image with energy and theta axes.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes. Default is None.
        fraction : float, optional
            Containment fraction between 0 and 1.
            Default is 0.68.
        add_cbar : bool, optional
            Add a colorbar. Default is True.
        axes_loc : dict, optional
            Keyword arguments passed to `~mpl_toolkits.axes_grid1.axes_divider.AxesDivider.append_axes`.
        kwargs_colorbar : dict, optional
            Keyword arguments passed to `~matplotlib.pyplot.colorbar`.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.pcolormesh`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
             Matplotlib axes.
        """
        ax = plt.gca() if ax is None else ax

        energy = self.axes["energy_true"]
        offset = self.axes["offset"]

        # Set up and compute data
        containment = self.containment_radius(
            energy_true=energy.center[:, np.newaxis],
            offset=offset.center,
            fraction=fraction,
        )

        # plotting defaults
        kwargs.setdefault("cmap", "GnBu")
        kwargs.setdefault("vmin", np.nanmin(containment.value))
        kwargs.setdefault("vmax", np.nanmax(containment.value))

        kwargs_colorbar = kwargs_colorbar or {}

        # Plotting
        with quantity_support():
            caxes = ax.pcolormesh(
                energy.edges, offset.edges, containment.value.T, **kwargs
            )

        energy.format_plot_xaxis(ax=ax)
        offset.format_plot_yaxis(ax=ax)

        if add_cbar:
            label = f"Containment radius R{100 * fraction:.0f} ({containment.unit})"
            kwargs_colorbar.setdefault("label", label)
            add_colorbar(caxes, ax=ax, axes_loc=axes_loc, **kwargs_colorbar)

        return ax

    def plot_psf_vs_rad(
        self, ax=None, offset=[0] * u.deg, energy_true=[0.1, 1, 10] * u.TeV, **kwargs
    ):
        """Plot PSF vs rad.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes. Default is None.
        offset : `~astropy.coordinates.Angle`, optional
            Offset in the field of view.
            Default is 0 deg.
        energy_true : `~astropy.units.Quantity`, optional
            True energy at which to plot the profile.
            Default is [0.1, 1, 10] TeV.
        kwargs : dict
            Keyword arguments.

        """
        from gammapy.datasets.map import RAD_AXIS_DEFAULT

        ax = plt.gca() if ax is None else ax

        try:
            rad = self.axes["rad"]
        except KeyError:
            rad = RAD_AXIS_DEFAULT

        with quantity_support():
            for theta in offset:
                for energy in energy_true:
                    psf_value = self.evaluate(
                        rad=rad.center, energy_true=energy, offset=theta
                    )
                    label = f"Offset: {theta:.1f}, Energy: {energy:.1f}"
                    ax.plot(rad.center, psf_value, label=label, **kwargs)

        rad.format_plot_xaxis(ax=ax)

        ax.set_yscale("log")
        ax.set_ylabel(f"PSF [{ax.yaxis.units.to_string(UNIT_STRING_FORMAT)}]")
        ax.legend()
        return ax

    def peek(self, figsize=(15, 5)):
        """Quick-look summary plots.

        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure. Default is (15, 5).

        """
        _, axes = plt.subplots(nrows=1, ncols=3, figsize=figsize)

        self.plot_containment_radius(fraction=0.68, ax=axes[0])
        self.plot_containment_radius(fraction=0.95, ax=axes[1])
        self.plot_containment_radius_vs_energy(ax=axes[2])
        plt.tight_layout()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from gammapy.utils.testing import requires_data, requires_dependency
import numpy as np
from gammapy.modeling.selection.inference_data import (
    inference_data_from_ultranest,
    generate_prior_samples,
    inference_data_from_sampler,
    resample_posterior,
)

from numpy.testing import assert_allclose
from gammapy.modeling.models import SkyModel
from gammapy.datasets import Datasets, SpectrumDatasetOnOff
from gammapy.modeling.sampler import Sampler
from gammapy.modeling.models import (
    UniformPrior,
    LogUniformPrior,
    PowerLawSpectralModel,
    Models,
)


@pytest.fixture
def datasets():
    datasets = Datasets()
    dataset = SpectrumDatasetOnOff.read(
        "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    )
    datasets.append(dataset)
    return datasets


@pytest.fixture
def ultranest_result(datasets):
    pwl1 = PowerLawSpectralModel(index=2.3)
    pwl1.amplitude.prior = LogUniformPrior(min=1e-12, max=1e-10)
    pwl1.index.prior = UniformPrior(min=2, max=3)

    models = Models([SkyModel(pwl1, name="source1")])
    datasets.models = models

    sampler_opts = {"live_points": 50}
    sampler = Sampler(backend="ultranest", sampler_opts=sampler_opts)

    ultranest_result = sampler.run(datasets)
    return ultranest_result


@requires_dependency("ultranest")
@requires_dependency("arviz")
@requires_data()
def test_inference_data_from_sampler_basic(ultranest_result, datasets):
    inference_data = inference_data_from_sampler(
        results=ultranest_result,
        datasets=datasets,
        backend="ultranest",
        n_posterior_samples=None,
        n_prior_samples=None,
        predictives=False,
    )

    assert "posterior" in inference_data.groups()
    assert "sample_stats" not in inference_data.groups()
    assert "prior" not in inference_data.groups()

    n_samples = len(ultranest_result.sampler_results["samples"][:, 0])
    assert inference_data.posterior.sizes["draw"] == n_samples

    with pytest.raises(ValueError):
        inference_data_from_sampler(
            results=ultranest_result,
            datasets=datasets,
            backend="notimplemented",
        )


@requires_dependency("ultranest")
@requires_dependency("arviz")
@requires_data()
def test_inference_data_from_sampler_with_options(ultranest_result, datasets):
    inference_data = inference_data_from_sampler(
        results=ultranest_result,
        datasets=datasets,
        backend="ultranest",
        n_posterior_samples=10,
        n_prior_samples=5,
        predictives=True,
    )

    assert "sample_stats" in inference_data.groups()
    assert "posterior" in inference_data.groups()
    assert "prior" in inference_data.groups()
    assert "posterior_predictive" in inference_data.groups()
    assert "prior_predictive" in inference_data.groups()
    assert "log_likelihood" in inference_data.groups()
    assert "log_prior" in inference_data.groups()
    assert "observed_data" in inference_data.groups()

    assert inference_data.posterior.sizes["draw"] == 10
    assert inference_data.log_likelihood.sizes["draw"] == 10
    assert inference_data.log_prior.sizes["draw"] == 10

    assert inference_data.prior.sizes["draw"] == 5
    assert inference_data.prior_predictive.sizes["draw"] == 5

    n_pixel = datasets[0].mask.data.sum()
    assert inference_data.observed_data.sizes["pixel"] == n_pixel
    assert inference_data.log_likelihood.sizes["pixel"] == n_pixel
    assert inference_data.posterior_predictive.sizes["pixel"] == n_pixel
    assert inference_data.prior_predictive.sizes["pixel"] == n_pixel


@requires_dependency("ultranest")
@requires_dependency("arviz")
@requires_data()
def test_resample_posterior(ultranest_result, datasets):
    inference_data = inference_data_from_sampler(
        results=ultranest_result,
        datasets=datasets,
        backend="ultranest",
        n_posterior_samples=20,
        n_prior_samples=5,
        predictives=False,
    )
    idata_resampled = resample_posterior(inference_data, n_samples=10)
    assert idata_resampled.posterior.sizes["draw"] == 10
    assert idata_resampled.prior.sizes["draw"] == 5


@requires_dependency("arviz")
def test_inference_data_from_ultranest():
    import arviz

    result_dict = {
        "paramnames": ["x", "y"],
        "samples": np.array([[1.0, 2.0], [3.0, 4.0]]),
        "weighted_samples": {
            "points": np.array([[1.0, 2.0], [3.0, 4.0]]),
            "upoints": np.array([[0.1, 0.2], [0.3, 0.4]]),
            "weights": np.array([0.4, 0.6]),
            "logl": np.array([-1.0, -0.5]),
        },
    }

    # unweighted
    data = inference_data_from_ultranest(result_dict, weighted=False)
    assert isinstance(data, arviz.InferenceData)
    assert "posterior" in data.groups()

    posterior = data.posterior
    assert set(posterior.data_vars) == {"x", "y"}
    assert_allclose(posterior.x.data.flatten(), result_dict["samples"][:, 0])
    assert_allclose(posterior.y.data.flatten(), result_dict["samples"][:, 1])

    # weighted
    data = inference_data_from_ultranest(result_dict, weighted=True)
    assert isinstance(data, arviz.InferenceData)
    assert "posterior" in data.groups()
    assert "sample_stats" in data.groups()
    assert "weights" in data.sample_stats
    assert "log_likelihood" in data.sample_stats

    posterior = data.posterior
    assert set(posterior.data_vars) == {"x", "y"}
    assert_allclose(
        posterior.x.data.flatten(), result_dict["weighted_samples"]["points"][:, 0]
    )
    assert_allclose(
        posterior.y.data.flatten(), result_dict["weighted_samples"]["points"][:, 1]
    )


def test_generate_prior_samples():
    class DummyPrior:
        def _inverse_cdf(self, val):
            return val * 10

    class DummyParameter:
        def __init__(self):
            self.prior = DummyPrior()

    parameters = [DummyParameter(), DummyParameter()]
    samples = generate_prior_samples(parameters, n_prior_samples=3, random_seed=0)
    assert samples.shape == (3, 2)
    assert np.all(samples >= 0) and np.all(samples <= 10)

    parameters[0].prior = None
    with pytest.raises(ValueError):
        generate_prior_samples(parameters, n_prior_samples=3, random_seed=0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Dark matter spectra."""

import numpy as np
import astropy.units as u
from astropy.table import Table
from gammapy.maps import Map, MapAxis, RegionGeom
from gammapy.modeling import Parameter
from gammapy.modeling.models import SpectralModel, TemplateNDSpectralModel
from gammapy.utils.scripts import make_path
from gammapy.utils.table import table_map_columns
import warnings

__all__ = [
    "PrimaryFlux",
    "DarkMatterAnnihilationSpectralModel",
    "DarkMatterDecaySpectralModel",
]


class PrimaryFlux(TemplateNDSpectralModel):
    """DM-annihilation gamma-ray spectra.

    Based on the precomputed models of PPPC4 DM ID by [1]_, [2]_ and CosmiXs by [3]_, [4]_.
    All available annihilation channels can be found there. The dark matter mass will be set
    to the nearest available value. The spectra will be available as
    `~gammapy.modeling.models.TemplateNDSpectralModel` for a chosen dark matter mass and
    annihilation channel. Using a `~gammapy.modeling.models.TemplateNDSpectralModel`
    allows the interpolation between different dark matter masses.

    Parameters
    ----------
    mDM : `~astropy.units.Quantity`
        Dark matter particle mass as rest mass energy.
    channel : str
        Annihilation channel. List available channels with `~gammapy.astro.darkmatter.PrimaryFlux.allowed_channels`.
    source : {"cosmixs", "pppc4"}, optional
        Data source for the spectra. Default is 'pppc4'.

    References
    ----------
    .. [1] `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection" <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    .. [2] `Cirelli et al. (2016), "PPPC 4 DM ID: A Poor Particle Physicist Cookbook for Dark Matter Indirect Detection" <http://www.marcocirelli.net/PPPC4DMID.html>`_
    .. [3] `Arina et al. (2024), "CosmiXs: Cosmic messenger spectra for indirect dark matter searches" <https://arxiv.org/abs/2312.01153>`_
    .. [4] `Di Mauro et al. (2025), "Nailing down the theoretical uncertainties of Dbar spectrum produced from dark matter" <https://arxiv.org/abs/2411.04815>`_

    """

    channel_registry = {
        "eL": "eL",
        "eR": "eR",
        "e": "e",
        "muL": r"\[Mu]L",
        "muR": r"\[Mu]R",
        "mu": r"\[Mu]",
        "tauL": r"\[Tau]L",
        "tauR": r"\[Tau]R",
        "tau": r"\[Tau]",
        "q": "q",
        "c": "c",
        "b": "b",
        "t": "t",
        "WL": "WL",
        "WT": "WT",
        "W": "W",
        "ZL": "ZL",
        "ZT": "ZT",
        "Z": "Z",
        "g": "g",
        "gamma": r"\[Gamma]",
        "h": "h",
        "nu_e": r"\[Nu]e",
        "nu_mu": r"\[Nu]\[Mu]",
        "nu_tau": r"\[Nu]\[Tau]",
        "V->e": "V->e",
        "V->mu": r"V->\[Mu]",
        "V->tau": r"V->\[Tau]",
        "aZ": "aZ",
        "HZ": "HZ",
        "d": "d",
        "u": "u",
        "s": "s",
    }

    mapping_dict_PPPC4_to_CosmiXs = {
        "DM": "mDM",
        "Log10[x]": "Log[10,x]",
        "dNdLog10x[eL]": "eL",
        "dNdLog10x[eR]": "eR",
        "dNdLog10x[e]": "e",
        "dNdLog10x[muL]": "\\[Mu]L",
        "dNdLog10x[muR]": "\\[Mu]R",
        "dNdLog10x[mu]": "\\[Mu]",
        "dNdLog10x[tauL]": "\\[Tau]L",
        "dNdLog10x[tauR]": "\\[Tau]R",
        "dNdLog10x[tau]": "\\[Tau]",
        "dNdLog10x[nue]": "\\[Nu]e",
        "dNdLog10x[numu]": "\\[Nu]\\[Mu]",
        "dNdLog10x[nutau]": "\\[Nu]\\[Tau]",
        "dNdLog10x[u]": "u",  # Does not exist explicitly on PPPC4, but it is equivalent to q
        "dNdLog10x[d]": "d",  # Does not exist explicitly on PPPC4, but it is equivalent to q
        "dNdLog10x[s]": "s",  # Does not exist explicitly on PPPC4, but it is equivalent to q
        "dNdLog10x[c]": "c",
        "dNdLog10x[b]": "b",
        "dNdLog10x[t]": "t",
        "dNdLog10x[a]": "\\[Gamma]",
        "dNdLog10x[g]": "g",
        "dNdLog10x[W]": "W",
        "dNdLog10x[WL]": "WL",
        "dNdLog10x[WT]": "WT",
        "dNdLog10x[Z]": "Z",
        "dNdLog10x[ZL]": "ZL",
        "dNdLog10x[ZT]": "ZT",
        "dNdLog10x[H]": "h",
        "dNdLog10x[aZ]": None,  # Does not exist  on PPPC4
        "dNdLog10x[HZ]": None,  # Does not exist  on PPPC4
    }

    tag = ["PrimaryFlux", "dm-pf"]

    def __init__(self, mDM, channel, source="pppc4"):
        if source is None:
            source = "pppc4"
            warnings.warn(
                "\nSince no spectra source has been chosen, PPPC4 will be used by default.\n",
                UserWarning,
            )
        self.source = source.lower()
        if self.source == "pppc4":
            table_filename = (
                "$GAMMAPY_DATA/dark_matter_spectra/PPPC4DMID/AtProduction_gammas.dat"
            )
        elif self.source == "cosmixs":
            table_filename = (
                "$GAMMAPY_DATA/dark_matter_spectra/cosmixs/AtProduction-Gamma.dat"
            )
        else:
            raise ValueError(
                "\n\nData source is not valid, please choose between PPPC4 or cosmixs\n"
            )

        self.table_path = make_path(table_filename)
        if not self.table_path.exists():
            raise FileNotFoundError(
                f"\n\nFile not found: {table_filename}\n"
                "You may download the dataset needed with the following command:\n"
                "gammapy download datasets --src dark_matter_spectra"
            )
        else:
            ascii_format = (
                "ascii.commented_header"
                if self.source == "cosmixs"
                else "ascii.fast_basic"
            )
            self.table = Table.read(
                str(self.table_path),
                format=ascii_format,
                guess=False,
                delimiter=" ",
            )
            if self.source == "cosmixs":
                self.table = table_map_columns(
                    self.table, self.mapping_dict_PPPC4_to_CosmiXs
                )

        self.channel = channel

        # create RegionNDMap for channel

        masses = np.unique(self.table["mDM"])
        log10x = np.unique(self.table["Log[10,x]"])

        mass_axis = MapAxis.from_nodes(masses, name="mass", interp="log", unit="GeV")
        log10x_axis = MapAxis.from_nodes(log10x, name="energy_true")

        channel_name = self.channel_registry[self.channel]

        geom = RegionGeom(region=None, axes=[log10x_axis, mass_axis])
        region_map = Map.from_geom(
            geom=geom, data=self.table[channel_name].reshape(geom.data_shape)
        )

        interp_kwargs = {"extrapolate": True, "fill_value": 0, "values_scale": "lin"}
        super().__init__(region_map, interp_kwargs=interp_kwargs)
        self.mDM = mDM
        self.mass.frozen = True

    @property
    def mDM(self):
        """Dark matter mass."""
        return u.Quantity(self.mass.value, "GeV")

    @mDM.setter
    def mDM(self, mDM):
        unit = self.mass.unit
        _mDM = u.Quantity(mDM).to(unit)
        _mDM_val = _mDM.to_value(unit)

        min_mass = u.Quantity(self.mass.min, unit)
        max_mass = u.Quantity(self.mass.max, unit)

        if _mDM_val < self.mass.min or _mDM_val > self.mass.max:
            raise ValueError(
                f"The mass {_mDM} is out of the bounds of the model. Please choose a mass between {min_mass} < `mDM` < {max_mass}"
            )

        self.mass.value = _mDM_val

    @property
    def allowed_channels(self):
        """List of allowed annihilation channels."""
        return list(self.channel_registry.keys())

    @property
    def channel(self):
        """Annihilation channel as a string."""
        return self._channel

    @channel.setter
    def channel(self, channel):
        if channel not in self.allowed_channels:
            raise ValueError(
                f"Invalid channel: {channel}\nAvailable: {self.allowed_channels}\n"
            )
        else:
            if self.source == "pppc4":
                if channel in ("aZ", "HZ"):
                    raise ValueError(
                        f"\n\nThe channel {channel} is not available in PPPC4, please choose another channel or use CosmiXs (cosmixs) as source\n"
                    )
                elif channel in ("d", "u", "s"):
                    raise ValueError(
                        f"\n\nThe channel {channel} is not available in PPPC4, please choose the equivalent channel q or use CosmiXs (cosmixs) as source\n"
                    )
                else:
                    self._channel = channel

            elif self.source == "cosmixs":
                if channel in ("V->e", "V->mu", "V->tau"):
                    raise ValueError(
                        f"\n\nThe channel {channel} is not available in CosmiXs, please choose another channel or use PPPC4 as source\n"
                    )
                elif channel == "q":
                    raise ValueError(
                        "\n\nThe channel q is not available in cosmixs, please choose an equivalent channel such as d, u or s or use PPPC4 as source\n"
                    )
                else:
                    self._channel = channel

    def evaluate(self, energy, *args):
        """Evaluate the primary flux."""
        args = list(args)
        args.append(self.mDM)

        log10x = np.log10(energy / self.mDM)

        dN_dlogx = super().evaluate(log10x, *args)
        dN_dE = dN_dlogx / (energy * np.log(10))
        return dN_dE


class DarkMatterAnnihilationSpectralModel(SpectralModel):
    r"""Dark matter annihilation spectral model.

    The gamma-ray flux is computed as follows:

    .. math::
        \frac{\mathrm d \phi}{\mathrm d E} =
        \frac{\langle \sigma\nu \rangle}{4\pi k m^2_{\mathrm{DM}}}
        \frac{\mathrm d N}{\mathrm dE} \times J(\Delta\Omega)

    Parameters
    ----------
    mass : `~astropy.units.Quantity`
        Dark matter mass.
    channel : str
        Annihilation channel for `~gammapy.astro.darkmatter.PrimaryFlux`, e.g. "b" for "bbar".
        See `PrimaryFlux.channel_registry` for more.
    scale : float
        Scale parameter for model fitting.
    jfactor : `~astropy.units.Quantity`
        Integrated J-Factor needed when `~gammapy.modeling.models.PointSpatialModel`
        is used.
    z : float, optional
        Redshift value. Default is 0.
    k : int, optional
        Type of dark matter particle (k:2 Majorana, k:4 Dirac). Default is 2.
    source : {"cosmixs", "pppc4"}, optional
        Data source for the spectra. Default is 'pppc4'.

    Examples
    --------
    This is how to instantiate a `DarkMatterAnnihilationSpectralModel` model::

        >>> import astropy.units as u
        >>> from gammapy.astro.darkmatter import DarkMatterAnnihilationSpectralModel

        >>> channel = "b"
        >>> massDM = 5000*u.Unit("GeV")
        >>> jfactor = 3.41e19 * u.Unit("GeV2 cm-5")
        >>> modelDM = DarkMatterAnnihilationSpectralModel(mass=massDM, channel=channel, jfactor=jfactor)  # noqa: E501

    References
    ----------
    `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection" <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    THERMAL_RELIC_CROSS_SECTION = 3e-26 * u.Unit("cm3 s-1")
    """Thermally averaged annihilation cross-section"""

    scale = Parameter(
        "scale",
        1,
        unit="",
        interp="log",
    )
    tag = ["DarkMatterAnnihilationSpectralModel", "dm-annihilation"]

    def __init__(
        self, mass, channel, scale=scale.quantity, jfactor=1, z=0, k=2, source="pppc4"
    ):
        self.k = k
        self.z = z
        self.mass = u.Quantity(mass)
        self.channel = channel
        self.jfactor = u.Quantity(jfactor)
        self.primary_flux = PrimaryFlux(mass, channel=self.channel, source=source)
        self.source = source
        super().__init__(scale=scale)

    def evaluate(self, energy, scale):
        """Evaluate dark matter annihilation model."""
        flux = (
            scale
            * self.jfactor
            * self.THERMAL_RELIC_CROSS_SECTION
            * self.primary_flux(energy=energy * (1 + self.z))
            / self.k
            / self.mass
            / self.mass
            / (4 * np.pi)
        )
        return flux

    def to_dict(self, full_output=False):
        """Convert to dictionary."""
        data = super().to_dict(full_output=full_output)
        data["spectral"]["channel"] = self.channel
        data["spectral"]["mass"] = self.mass.to_string()
        data["spectral"]["jfactor"] = self.jfactor.to_string()
        data["spectral"]["z"] = self.z
        data["spectral"]["k"] = self.k
        data["spectral"]["source"] = self.source
        return data

    @classmethod
    def from_dict(cls, data):
        """Create spectral model from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary with model data.

        Returns
        -------
        model : `DarkMatterAnnihilationSpectralModel`
            Dark matter annihilation spectral model.
        """
        data = data["spectral"]
        data.pop("type")
        parameters = data.pop("parameters")
        scale = [p["value"] for p in parameters if p["name"] == "scale"][0]
        return cls(scale=scale, **data)


class DarkMatterDecaySpectralModel(SpectralModel):
    r"""Dark matter decay spectral model.

    The gamma-ray flux is computed as follows:

    .. math::
        \frac{\mathrm d \phi}{\mathrm d E} =
        \frac{\Gamma}{4\pi m_{\mathrm{DM}}}
        \frac{\mathrm d N}{\mathrm dE} \times J(\Delta\Omega)

    Parameters
    ----------
    mass : `~astropy.units.Quantity`
        Dark matter mass.
    channel : str
        Decay channel for `~gammapy.astro.darkmatter.PrimaryFlux`, e.g. "b" for "bbar".
        See `PrimaryFlux.channel_registry` for more.
    scale : float
        Scale parameter for model fitting
    jfactor : `~astropy.units.Quantity`
        Integrated J-Factor needed when `~gammapy.modeling.models.PointSpatialModel`
        is used.
    z : float, optional
        Redshift value. Default is 0.
    source : {"cosmixs", "pppc4"}, optional
        Data source for the spectra. Default is 'pppc4'.

    Examples
    --------
    This is how to instantiate a `DarkMatterDecaySpectralModel` model::

        >>> import astropy.units as u
        >>> from gammapy.astro.darkmatter import DarkMatterDecaySpectralModel

        >>> channel = "b"
        >>> massDM = 5000*u.Unit("GeV")
        >>> jfactor = 3.41e19 * u.Unit("GeV cm-2")
        >>> modelDM = DarkMatterDecaySpectralModel(mass=massDM, channel=channel, jfactor=jfactor)  # noqa: E501

    References
    ----------
    `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection" <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    LIFETIME_AGE_OF_UNIVERSE = 4.3e17 * u.Unit("s")
    """Use age of univserse as lifetime"""

    scale = Parameter(
        "scale",
        1,
        unit="",
        interp="log",
    )

    tag = ["DarkMatterDecaySpectralModel", "dm-decay"]

    def __init__(
        self, mass, channel, scale=scale.quantity, jfactor=1, z=0, source="pppc4"
    ):
        self.z = z
        self.mass = u.Quantity(mass)
        self.channel = channel
        self.jfactor = u.Quantity(jfactor)
        self.primary_flux = PrimaryFlux(
            self.mass / 2, channel=self.channel, source=source
        )
        self.source = source
        super().__init__(scale=scale)

    def evaluate(self, energy, scale):
        """Evaluate dark matter decay model."""
        flux = (
            scale
            * self.jfactor
            * self.primary_flux(energy=energy * (1 + self.z))
            / self.LIFETIME_AGE_OF_UNIVERSE
            / self.mass
            / (4 * np.pi)
        )
        return flux

    def to_dict(self, full_output=False):
        data = super().to_dict(full_output=full_output)
        data["spectral"]["channel"] = self.channel
        data["spectral"]["mass"] = self.mass.to_string()
        data["spectral"]["jfactor"] = self.jfactor.to_string()
        data["spectral"]["z"] = self.z
        data["spectral"]["source"] = self.source
        return data

    @classmethod
    def from_dict(cls, data):
        """Create spectral model from dictionary.

        Parameters
        ----------
        data : dict
            Dictionary with model data.

        Returns
        -------
        model : `DarkMatterDecaySpectralModel`
            Dark matter decay spectral model.
        """
        data = data["spectral"]
        data.pop("type")
        parameters = data.pop("parameters")
        scale = [p["value"] for p in parameters if p["name"] == "scale"][0]
        return cls(scale=scale, **data)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from gammapy.modeling.parameter import Parameters


def inference_data_from_ultranest(sampler_results, weighted=False):
    """
     Convert UltraNest result dictionary to ArviZ InferenceData.

     Parameters
     ----------
     sampler_results : dict
         The result dictionary returned by `ReactiveNestedSampler.run()`.
     weighted : bool
         If True, uses the weighted samples (more accurate) otherwise use directly the unweighted samples.

     Returns
     -------
    inferencedata : `arviz.InferenceData`
        Returns an arviz.InferenceData instance
    """

    import arviz as az
    import xarray as xr

    var_names = sampler_results["paramnames"]

    if weighted:
        ws = sampler_results["weighted_samples"]

        # Posterior samples (1 chain)
        posterior_data = {
            var: (["chain", "draw"], ws["points"][:, i][np.newaxis, :])
            for i, var in enumerate(var_names)
        }
        posterior = xr.Dataset(posterior_data)

        unconstrained_posterior_data = {
            var: (["chain", "draw"], ws["upoints"][:, i][np.newaxis, :])
            for i, var in enumerate(var_names)
        }
        unconstrained_posterior = xr.Dataset(unconstrained_posterior_data)

        # Sample stats
        sample_stats = xr.Dataset(
            {
                "weights": (["chain", "draw"], ws["weights"][np.newaxis, :]),
                "log_likelihood": (["chain", "draw"], ws["logl"][np.newaxis, :]),
            }
        )

        return az.InferenceData(
            posterior=posterior,
            unconstrained_posterior=unconstrained_posterior,
            sample_stats=sample_stats,
        )
    else:
        ws = sampler_results["samples"]

        posterior_data = {
            var: (["chain", "draw"], ws[:, i][np.newaxis, :])
            for i, var in enumerate(var_names)
        }
        posterior = xr.Dataset(posterior_data)

        return az.InferenceData(posterior=posterior)


def inference_data_from_sampler(
    results,
    datasets,
    backend="ultranest",
    n_posterior_samples=None,
    n_prior_samples=None,
    random_seed=42,
    predictives=True,
):
    """
    Convert Sampler results to an ArviZ InferenceData object with optional resampling and prior inclusion.

    Parameters
    ----------
    results : `~gammapy.modeling.SamplerResult`
        The sampler results and model information.
    datasets: `gammapy.datasets.Datasets`
        Datasets that were used to obtain the results.
    backend : {"ultranest"}
        Global backend used for sampler. Default is "ultranest".
        UltraNest: Most options can be found in the
        `UltraNest doc <https://johannesbuchner.github.io/UltraNest/>`__.
    n_posterior_samples : int, optional
        Number of samples to generate after resampling the posterior to take into account weights.
        Default is None, which use the unweighted samples from ultranest.
     n_prior_samples : int, optional
        If provided, generates this number of samples from the prior distribution using the model's
        prior transform and include them in the 'prior' group of the InferenceData.
    random_seed : int, optional
        Seed for reproducibility when resampling posterior or generating prior samples. Default is 42.
    predictives : bool, optional
        If True, computes predicted counts and pointwise likelihood matrix. Default is True.

    Returns
    -------
    inference_data : `arviz.InferenceData`
        An InferenceData object containing posterior samples, optionally resampled,
        prior samples (if requested), and log-evidence attributes ('logz' and 'logzerr').
    """
    if backend == "ultranest":
        inference_data_constructor = inference_data_from_ultranest
    else:
        raise ValueError(f"Only ultranest backend is supported, got {backend}")

    sampler_results = results.sampler_results
    if n_posterior_samples is None:
        inference_data = inference_data_constructor(sampler_results)
    else:
        inference_data = inference_data_constructor(sampler_results, weighted=True)
        inference_data = resample_posterior(
            inference_data, n_samples=n_posterior_samples, random_seed=random_seed
        )  # unweighting

    # Add prior  group
    if n_prior_samples is not None:
        add_prior_samples(inference_data, results, n_prior_samples, random_seed)
        if predictives:
            add_sample_wise_quantities(inference_data, datasets, "prior")

    # Add posterior and log likelihood groups
    if predictives:
        add_sample_wise_quantities(inference_data, datasets, "posterior")
    return inference_data


def resample_posterior(inference_data, n_samples=None, random_seed=42):
    """
    Resample posterior samples from an InferenceData object using importance weights.

    Parameters
    ----------
    inference_data : `arviz.InferenceData`
        The input InferenceData object with posterior and sample_stats["weights"].
    n_samples : int, optional
        Number of resampled draws. If None, uses the number of original draws.
    random_seed : int, optional
        Random seed for reproducibility. Default is 42.

    Returns
    -------
    inferencedata : `arviz.InferenceData`
        A new InferenceData object with resampled posterior and sample_stats.
    """

    import arviz as az
    import xarray as xr

    rng = np.random.default_rng(random_seed)

    # Extract posterior samples and weights
    posterior = inference_data.posterior
    weights = inference_data.sample_stats["weights"].values[0]  # shape: (draw,)
    weights = weights / weights.sum()  # normalize

    n_draws = posterior.sizes["draw"]
    n_samples = n_samples or n_draws

    # Resample indices
    resampled_idx = rng.choice(n_draws, size=n_samples, replace=True, p=weights)

    # Resample each group
    resampled_groups = {}
    for group in inference_data.groups():
        dataset = getattr(inference_data, group)
        if group in ["prior", "prior_predictive", "observed_data"]:
            resampled_groups[group] = dataset
        else:
            resampled_data = {}
            for var in dataset.data_vars:
                values = dataset[var].values
                # Resample along draw dimension (axis=1)
                resampled_values = values[:, resampled_idx]
                resampled_data[var] = (["chain", "draw"], resampled_values)
            resampled_groups[group] = xr.Dataset(resampled_data)

    # Create new InferenceData with resampled groups
    return az.InferenceData(**resampled_groups)


def generate_prior_samples(parameters, n_prior_samples=1000, random_seed=42):
    """Generate prior samples. This function draws samples from the prior distributions of the model
    parameters using inverse transform sampling.

    Parameters
    ----------
    parameters : `~gammapy.modeling.Parameters` or list of `~gammapy.modeling.Parameter`
        The list of model parameters with priors to sample.
    n_prior_samples : int, optional
        Number of prior samples to generate. If None, no samples are added.
    random_seed : int, optional
        Seed for the random number generator to ensure reproducibility. Default is 42.

    Returns
    -------
    prior_samples : `numpy.array`
        An array with dimension `n_prior_samples` times len(parameters).
    """

    def _prior_inverse_cdf(values):
        """Returns a list of model parameters for a given list of values (that are bound in [0,1])."""
        if None in Parameters(parameters).prior:
            raise ValueError(
                "Some parameters have no prior set. You need priors on all parameters."
            )
        return [par.prior._inverse_cdf(val) for par, val in zip(parameters, values)]

    # Generate uniform samples in unit cube
    rng = np.random.default_rng(random_seed)
    unit_cube_samples = rng.uniform(size=(n_prior_samples, len(parameters)))

    # Transform to physical space using the prior transform
    prior_samples = np.array([_prior_inverse_cdf(u) for u in unit_cube_samples])
    return prior_samples


def add_prior_samples(inference_data, results, n_prior_samples=None, random_seed=42):
    """
    Generate and add prior samples to an ArviZ InferenceData object.

    This function draws samples from the prior distributions of the model
    parameters using inverse transform sampling. The samples are added to
    the `prior` group of the provided `InferenceData` object.

    Parameters
    ----------
    inference_data : `arviz.InferenceData`
        The inference data object to which the prior samples will be added.
    results : `SamplerResult`
        The sampler result object containing the model and its priors.
    n_prior_samples : int, optional
        Number of prior samples to generate. If None, no samples are added.
    random_seed : int, optional
        Seed for the random number generator to ensure reproducibility. Default is 42.

    Raises
    ------
    `ValueError`
        If any model parameter does not have a prior defined.
    """
    import xarray as xr

    parameters = results.models.parameters.free_unique_parameters
    prior_samples = generate_prior_samples(parameters, n_prior_samples, random_seed)

    # Create prior group
    prior = xr.Dataset(
        {
            name: (["chain", "draw"], prior_samples[:, i][np.newaxis, :])
            for i, name in enumerate(parameters.names)
        }
    )
    inference_data.add_groups(prior=prior)


def add_sample_wise_quantities(inference_data, datasets, group_name):
    """
    Compute sample-wise quantities such as log-likelihood, log-prior, and predicted counts.

    This function evaluates the model for each sample in the specified group
    (posterior or prior) and computes the corresponding log-likelihood,
    log-prior, and predicted counts. These are added to the `InferenceData`
    object under appropriate groups.

    Parameters
    ----------
    inference_data : `arviz.InferenceData`
        The inference data object containing posterior or prior samples.
    datasets : `Datasets`
        The datasets used for computing likelihood and predictions.
    group_name : {"posterior", "prior"}
        The group of samples to evaluate.

    Notes
    -----
    - Assumes that the `stat_array()` method returns `-2 * log-likelihood`.
    - The function modifies the `InferenceData` object in-place.
    """
    import xarray as xr

    group = getattr(inference_data, group_name)
    n_samples = group.sizes["draw"]
    log_likelihood_matrix = []
    log_prior_matrix = []
    npred_matrix = []
    with datasets.models.parameters.restore_status():
        for idx in range(
            n_samples
        ):  # this should be vectorialized, or parallelize or gammapy should dump stat_array on disk
            for p in datasets.models.parameters.free_parameters:
                p.value = group[p.name][0, idx].item()

            if group_name == "posterior":
                stat_array = (
                    np.hstack([d.stat_array()[d.mask] for d in datasets]) / -2.0
                )  # assuming we have stat as -2lnL
                prior_stat_sum = np.sum(
                    [
                        p.prior._random_variable.logpdf(p.value)
                        for p in datasets.models.parameters.free_unique_parameters
                    ]
                )
                log_likelihood_matrix.append(stat_array)
                log_prior_matrix.append(prior_stat_sum)
            npred = np.hstack([d.npred().data[d.mask] for d in datasets])
            npred_matrix.append(npred)

        if group_name == "posterior":
            log_likelihood_matrix = np.array(log_likelihood_matrix)[np.newaxis, :, :]
            log_prior_matrix = np.array(log_prior_matrix)[np.newaxis, :]
            counts = np.hstack([d.counts.data[d.mask] for d in datasets])
        npred_matrix = np.array(npred_matrix)[np.newaxis, :, :]

    if group_name == "posterior":
        log_ligkelihood = xr.Dataset(
            {"log_ligkelihood": (["chain", "draw", "pixel"], log_likelihood_matrix)}
        )
        inference_data.add_groups(log_likelihood=log_ligkelihood)

        log_prior = xr.Dataset({"log_prior": (["chain", "draw"], log_prior_matrix)})
        inference_data.add_groups(log_prior=log_prior)

        posterior_predictive = xr.Dataset(
            {"counts": (["chain", "draw", "pixel"], npred_matrix)}
        )
        inference_data.add_groups(posterior_predictive=posterior_predictive)

        observed_data = xr.Dataset({"counts": (["pixel"], counts)})
        inference_data.add_groups(observed_data=observed_data)
    elif group_name == "prior":
        prior_predictive = xr.Dataset(
            {"counts": (["chain", "draw", "pixel"], npred_matrix)}
        )
        inference_data.add_groups(prior_predictive=prior_predictive)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import re
import numpy as np
from astropy.table import Table
from astropy.time import Time
import click
from github import Github, GithubException

log = logging.getLogger(__name__)


class GitHubContributorsExtractor:
    """Class to interact with GitHub and extract PR and issues info tables.

    Parameters
    ----------
    repo : str
        input repository. Default is 'gammapy/gammapy'
    token : str
        GitHub access token. Default is None
    """

    def __init__(self, repo=None, token=None):
        self.repo = repo if repo else "gammapy/gammapy"
        self.github = self.login(token)
        self.repo = self.github.get_repo(repo)

    @staticmethod
    def login(token):
        if token:
            g = Github(token, per_page=100)
        else:
            g = Github()

        try:
            user_login = g.get_user().login
        except GithubException:
            user_login = "anonymous"

        log.info(f"Logging in GitHub as {user_login}")
        return g

    def check_requests_number(self):
        remaining, total = self.github.rate_limiting
        log.info(f"Remaining {remaining} requests over {total} requests.")

    def extract_contributors_by_milestone(
        self, milestone_name, state="closed",
    ):
        """Extract list of unique contributors from PRs as per the milestone.

         Parameters
         ----------
         milestone_name :  str
            Milestone name i.e. 'v1.0'
         state : str ("closed", "open", "all")
            State of PRs to extract.
        """
        milestones = self.repo.get_milestones(state="all")
        milestone_obj = None
        for m in milestones:
            if m.title == milestone_name:
                milestone_obj = m
                break
        if milestone_obj is None:
            log.error(f"Milestone '{milestone_name}' not found in repository '{self.repo.full_name}'.")
            return []

        # Get PRs filtered by milestone using issues API (GitHub returns PRs as issues)
        issues = self.repo.get_issues(state=state, milestone=milestone_obj)

        self.check_requests_number()

        unique_users = set()
        num_prs = 0
        num_closed_issues = 0
        for issue in issues:
            if issue.pull_request is None:
                if issue.state == 'closed':
                    num_closed_issues += 1
                continue  # skip issues, only process PRs

            try:
                pr = self.repo.get_pull(issue.number)
            # Sometimes PRs are marked that way but cannot be found, because they no longer exist.
            except GithubException:
                continue

            # Skip PRs that are closed and not merged
            if not pr.merged:
                continue

            if "Backport" in pr.title:
                continue
            num_prs += 1
            log.info(f"Extracting Pull Request {pr.number}.")

            # It is possible that there will be 'authors' such as 'web-flow' which is just a merging action
            # The extra lines here ignore those users that are not real
            # Start to add authors
            if pr.user and pr.user.login not in {"web-flow", "dependabot[bot]", "github-actions[bot]"}:
                unique_users.add(pr.user.name or pr.user.login)

            # For committers
            for commit in pr.get_commits():
                if commit.committer and commit.committer.login not in {"web-flow", "dependabot[bot]",
                                                                       "github-actions[bot]"}:
                    # This is required for if a user deletes their account
                    try:
                        name = commit.committer.name
                        login = commit.committer.login
                    except GithubException:
                        continue
                    unique_users.add(name or login)

            # For reviewers
            for review in pr.get_reviews():
                if review.user and review.user.login not in {"web-flow", "dependabot[bot]", "github-actions[bot]"}:
                    unique_users.add(review.user.name or review.user.login)

        return {
            "contributors": sorted(list(unique_users)),
            "num_prs": num_prs,
            "num_closed_issues": num_closed_issues,
        }


@click.group()
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
)
def cli(log_level):
    logging.basicConfig(level=log_level)
    log.setLevel(level=log_level)

@cli.command(
    "contributors_by_milestone",
    help="Make a list of contributors for a specific milestone"
)
@click.option("--token", default=None, type=str, help="Your GitHub token.")
@click.option("--repo", default="gammapy/gammapy", type=str, help="The relative repo.")
@click.option("--milestone", required=True, type=str, help="Comma-separated list of milestones, e.g., '2.0.1,2.1'")
@click.option("--state", default="closed", type=str, help="Is the issues closed or not.")
def contributors_by_milestone(repo, token, milestone, state):
    """List contributors attached to a specific milestone."""
    extractor = GitHubContributorsExtractor(repo=repo, token=token)
    milestone_list = [m.strip() for m in milestone.split(",") if m.strip()]

    all_users = set()
    for m in milestone_list:
        log.info(f"Making list of contributors for milestone '{m}'.")
        info = extractor.extract_contributors_by_milestone(
            milestone_name=m,
            state=state,
        )
        users = info['contributors']
        num_prs = info['num_prs']
        num_closed_issues = info['num_closed_issues']
        log.info(f"""
        For milestone '{m}':
          - {num_prs} pull requests
          - {num_closed_issues} closed issues
          - {len(users)} unique contributors
        """)
        all_users.update(users)

    print(f"\nContributors for milestone '{milestone}'\n{'~' * 20}")
    for user in sorted(all_users):
        print(f"- {user}")

    # Add the names directly to the changelog
    changelog_file = "docs/release-notes/CHANGELOG.rst"
    contributors_sorted = sorted(all_users)
    with open(changelog_file, "a", encoding="utf-8") as f:
        for name in contributors_sorted:
            f.write(f"- {name}\n")


if __name__ == "__main__":
    cli()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest
import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u

from gammapy.visualization.violin import plot_samples_violin_vs_energy


def test_plot_samples_violin_vs_energy_invalid():
    with pytest.raises(ValueError, match="samples_per_band must match number of bins"):
        energy_edges = np.array([1, 2, 4]) * u.TeV

        samples_per_band = [
            np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
        ]

        _, ax = plt.subplots()
        plot_samples_violin_vs_energy(
            ax=ax,
            energy_edges=energy_edges,
            samples_per_band=samples_per_band,
        )

    with pytest.raises(ValueError, match="weights_per_band must match number of bins."):
        energy_edges = np.array([1, 2, 4]) * u.TeV

        samples_per_band = [
            np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
            np.array([0.4, 0.6, 0.5]) * u.Unit("cm-2 s-1 TeV-1"),
        ]
        weights_per_band = [
            np.array([1.0, 2.0, 1.0]),
        ]

        _, ax = plt.subplots()
        plot_samples_violin_vs_energy(
            ax=ax,
            energy_edges=energy_edges,
            samples_per_band=samples_per_band,
            weights_per_band=weights_per_band,
        )

    with pytest.raises(
        ValueError, match="energy_edges must be 1D and contain at least two values."
    ):
        energy_edges = np.array([1]) * u.TeV

        samples_per_band = [
            np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
            np.array([0.4, 0.6, 0.5]) * u.Unit("cm-2 s-1 TeV-1"),
        ]

        _, ax = plt.subplots()
        plot_samples_violin_vs_energy(
            ax=ax,
            energy_edges=energy_edges,
            samples_per_band=samples_per_band,
        )

    with pytest.raises(
        ValueError, match="energy_edges must be strictly increasing, finite, and > 0."
    ):
        energy_edges = np.array([-np.inf, 2, 3]) * u.TeV

        samples_per_band = [
            np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
            np.array([0.4, 0.6, 0.5]) * u.Unit("cm-2 s-1 TeV-1"),
        ]

        _, ax = plt.subplots()
        plot_samples_violin_vs_energy(
            ax=ax,
            energy_edges=energy_edges,
            samples_per_band=samples_per_band,
        )

    with pytest.raises(
        ValueError, match="violin_clip must satisfy 0 <= low < high <= 1."
    ):
        energy_edges = np.array([1, 2, 3]) * u.TeV

        samples_per_band = [
            np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
            np.array([0.4, 0.6, 0.5]) * u.Unit("cm-2 s-1 TeV-1"),
        ]

        _, ax = plt.subplots()
        plot_samples_violin_vs_energy(
            ax=ax,
            energy_edges=energy_edges,
            samples_per_band=samples_per_band,
            violin_clip=[10, 20],
        )


def test_plot_samples_violin_vs_energy_minimal():
    """Test that plot_samples_violin_vs_energy runs and returns artists for a simple case."""

    energy_edges = np.array([1, 2, 4]) * u.TeV

    samples_per_band = [
        np.array([1.0, 1.2, 0.8]) * u.Unit("cm-2 s-1 TeV-1"),
        np.array([-0.4, -0.6, 0.5]) * u.Unit("cm-2 s-1 TeV-1"),
    ]

    weights_per_band = [
        np.array([1.0, 2.0, 1.0]),
        None,  # second bin: unweighted
    ]

    fig, ax = plt.subplots()

    ax = plot_samples_violin_vs_energy(
        ax=ax,
        energy_edges=energy_edges,
        samples_per_band=samples_per_band,
        weights_per_band=weights_per_band,
        bw_method="scott",
        energy_power=0.0,
        grid_size=50,
        violin_clip=(0.05, 0.95),
        y_label="dN/dE",
    )

    assert len(ax.lines) > 0

    assert ax.get_xscale() == "log"
    assert ax.get_yscale() == "log"

    plt.close(fig)


def test_plot_samples_violin_vs_energy_empty_weights_ok():
    """Test that bins with empty or zero-sum weights are skipped safely."""

    energy_edges = np.array([1, 3]) * u.TeV
    samples_per_band = [np.array([1.0]) * u.Unit("cm-2 s-1 TeV-1")]
    weights_per_band = [np.array([0.0])]  # zero-sum → skipped internally

    fig, ax = plt.subplots()

    ax = plot_samples_violin_vs_energy(
        ax=ax,
        energy_edges=energy_edges,
        samples_per_band=samples_per_band,
        weights_per_band=weights_per_band,
    )

    assert len(ax.lines) == 0
    plt.close(fig)


def test_plot_samples_violin_vs_energy_no_weights():
    """Test that plot_samples_violin_vs_energy works when no weights are provided (uniform implied)."""

    energy_edges = np.array([1.0, 2.0, 4.0]) * u.TeV

    flux_unit = u.Unit("cm-2 s-1 TeV-1")
    samples_per_band = [
        np.array([1.0, 1.2, 0.8, 1.1]) * flux_unit,  # bin 1
        np.array([0.5, 0.6, 0.55, 0.52]) * flux_unit,  # bin 2
    ]

    fig, ax = plt.subplots()
    violin_kwargs = dict(
        color="C0",
        alpha=0.4,
        edgecolor="black",
        lw=1.0,
    )
    ax = plot_samples_violin_vs_energy(
        ax=ax,
        energy_edges=energy_edges,
        samples_per_band=samples_per_band,
        weights_per_band=None,
        energy_power=2,
        bw_method="scott",
        grid_size=64,
        violin_kwargs=violin_kwargs,
        y_label="dN/dE",
    )

    assert len(ax.lines) > 0

    assert ax.get_xscale() == "log"
    assert ax.get_yscale() == "log"

    assert "Energy" in ax.get_xlabel()
    assert "dN/dE" in ax.get_ylabel()

    plt.close(fig)
